/*  SUBROUTINE FOR BOND CALCULATIONS - CALLS USER SPECIFEID FUNCTION */
/*  FOR EACH ATOM PAIR BETWEEN rmin AND rmax                         */
/*  FUNCTION HAS FORM  void fbnd (int i, int j, float r)             */
/*  WHERE i,j ARE ATOM INDICES AND r IS INTERATOMIC DISTANCE         */


/*
************************************************************************
Includes
************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <bondsub.h>

/*
************************************************************************
Defines
************************************************************************
*/
#define UCHAR unsigned char

#define NDIR 3
#define X 0
#define Y 1
#define Z 2


/*
************************************************************************
Local Function Prototypes
************************************************************************
*/
void __sort (unsigned *, unsigned *, unsigned *, unsigned *, int, int);
int  GetTypePattern (int, int);
void TypeRDF_func (int i, int j, double rsq);
void RDF_func     (int i, int j, double rsq);



/*
************************************************************************
Module Variables
************************************************************************
*/
static long   *gl_table;
static double  gl_rmin;
static double  gl_rsqmin;
static double  gl_rsqmax;
static double  gl_fac;
static UCHAR  *gl_TypeList;
static int     gl_TypePattern;

/*
************************************************************************
Exported Functions
************************************************************************
*/

void bond (float **ct, unsigned n, float rmin, float rmax,
           float *box, unsigned *rep, void (*fbnd)(int,int,float)  )
{
   float           *cx,*cy,*cz;
   unsigned        *ci,*cj,*ck, *ind;
   unsigned        i,j,k, intv;
   unsigned        hor,ver,dep, xs,ys,zs, xm;
   unsigned        ran1,ran2,ran3, bi,bj,bk, b2i,b2j,b2k, ranmin;
   char            rx,ry,rz, cont;
   float           bx,by,bz, *u,umin,umax;
   float           min[3], max[3], ran[3], maxbox, scale, iscale2;
   float           rs,tx,ty,tz,ranminsq, ranmaxsq;
   /*  FOR 2 byte unsigned, maxint = 0x3FFFF    */
   /*      4 byte unsigned, maxint = 0x3fffffff */
   unsigned        maxint = (unsigned) ~0 >> 2;

   if ( (n==0) || ( (rmin==0.0) && (rmax==0.0) ) )
      return;

   /*   FIND DIMENSION WITH MAXIMUM EXTENT    */
   for (i=0;i<3;i++) {
      u    = ct[i];
      umax = umin = ct[i][0];
      for (j=0;j<n;j++)
         if (u[j]<umin)
            umin = u[j];
         else if (u[j]>umax)
            umax = u[j];
      min[i] = umin;
      max[i] = umax;
      ran[i] = umax-umin;
   }

   /*  ORDER DIMENSIONS BY EXTENT  */
   cont = 1;
   for (i=0;i<3 && cont;i++) {
      for (j=(i+1)%3;j!=i && cont;j=(j+1)%3) {
         k = 3-i-j;
         if (ran[i]>=ran[j] && ran[i]>=ran[k] && ran[j]>=ran[k]) {
            hor = i;
            ver = j;
            dep = k;
            cont = 0;
         }
      }
   }

   cx  = ct[hor];
   cy  = ct[ver];
   cz  = ct[dep];
   rx  = rep[hor];
   ry  = rep[ver];
   rz  = rep[dep];
   bx  = box[hor];
   by  = box[ver];
   bz  = box[dep];
   /* SCALE DOUBLE TO UNSIGNED  */
   maxbox = bx;
   if (by>maxbox)
      maxbox = by;
   if (bz>maxbox)
      maxbox = bz;

   /*  SCALE COORDINATES TO UNSIGNED INTEGER VALUES  */
   scale = maxint / maxbox;
   iscale2 = 1.0/(scale*scale);
   ci = (unsigned *) calloc (n,sizeof(unsigned));
   cj = (unsigned *) calloc (n,sizeof(unsigned));
   ck = (unsigned *) calloc (n,sizeof(unsigned));
   ind= (unsigned *) calloc (n,sizeof(unsigned));
   if (ind==NULL) {
      printf ("   ERROR - Not enough memory to calculate neighbor list.\n");
      exit(1);
   }

   for (i=0;i<n;i++) {
      ind[i] = i;
      ci[i] = scale * (cx[i] - min[X]);
      cj[i] = scale * (cy[i] - min[Y]);
      ck[i] = scale * (cz[i] - min[Z]);
   }

   bi = scale * bx;
   bj = scale * by;
   bk = scale * bz;
   b2i = scale * (bx - rmax) - 1;
   b2j = scale * (by - rmax) - 1;
   b2k = scale * (bz - rmax) - 1;
   /*  SORT BY CURRENT X DIRECTION  */
   __sort (ci,cj,ck, ind, 0, n-1);
   /*  CONVERT CONSTANTS TO INTEGER  */
   ran1   = scale*rmax*sqrt(1) + 1;
   ran2   = scale*rmax*sqrt(2) + 1;
   ran3   = scale*rmax*sqrt(3) + 1;
   ranmin = scale*rmin         + 1;
   ranminsq = (tx=ranmin)*tx;    /*  These two variables are float  */
   ranmaxsq = (tx=ran1  )*tx;
   /*  FIND START OF ELIGILITY INTERVAL  */
   if (rx && 2*rmax<bi) {
      xm = bi + ci[0] - ran1;
      intv = n;
      while (--intv>0 && ci[intv]>xm);
      intv++;
   }
   else
      intv = 0;
   /*  FIND NEIGHBORS  */
   for (i=0;i<n;i++) {
      /*  need xj every iteration so calculation bj incrementing  */
      /*  yj and zj needed less frequently so do index            */
      if (intv==n)
         intv = 0;
      j=intv;
      while (j!=i) {
         xs = abs(ci[i]-ci[j]);
         if (rx && xs>b2i)
            xs = bi-xs;
         if (xs>=ran1)
            intv = j+1;
         else {
            ys = abs(cj[i]-cj[j]);
            if (ry && ys>b2j)
               ys = bj-ys;
            if (ys<ran1 && xs+ys<ran2) {
               zs = abs(ck[i]-ck[j]);
               if (rz && zs>b2k)
                  zs = bk-zs;
               if (zs<ran1 && xs+ys+zs<ran3) {
                  rs = (tx=xs)*tx + (ty=ys)*ty + (tz=zs)*tz;
                  if (rs>=ranminsq && rs<ranmaxsq)
                     fbnd (ind[i],ind[j],rs*iscale2);
               }
            }
         }
         if (++j==n)
            j=0;
      }
   }
   /*  RETURN  STORAGE  */
   free (ind);
   free (ck);
   free (cj);
   free (ci);
}

/*  THIS FUNCTION EXPECTS COORDINATES IN SINGLE ARRAY OF x,y,z,x,y,z .. */
void bondd (double *c, unsigned n, double rmin, double rmax,
            double *box, unsigned *rep, void (*fbnd)(int,int,double)  )
{
   unsigned        *ci,*cj,*ck, *ind;
   unsigned        i,j,k, intv;
   unsigned        hor,ver,dep, xs,ys,zs, xm;
   unsigned        ran1,ran2,ran3, bi,bj,bk, b2i,b2j,b2k, ranmin;
   char            rx,ry,rz, cont;
   double           bx,by,bz, umin,umax;
   double           min[3], max[3], ran[3], maxbox, scale, iscale2;
   double           rs,tx,ty,tz,ranminsq, ranmaxsq;
   /*  FOR 2 byte unsigned, maxint = 0x3FFFF    */
   /*      4 byte unsigned, maxint = 0x3fffffff */
   unsigned        maxint = (unsigned) ~0 >> 2;

   if ( (n==0) || ( (rmin==0.0) && (rmax==0.0) ) )
      return;

   /*   FIND DIMENSION WITH MAXIMUM EXTENT    */
   for (i=0;i<3;i++) {
      umax = umin = c[i];
      for (j=i;j<3*n;j+=3)
         if (c[j]<umin)
            umin = c[j];
         else if (c[j]>umax)
            umax = c[j];
      min[i] = umin;
      max[i] = umax;
      ran[i] = umax-umin;
   }

   /*  ORDER DIMENSIONS BY EXTENT  */
   cont = 1;
   for (i=0;i<3 && cont;i++) {
      for (j=(i+1)%3;j!=i && cont;j=(j+1)%3) {
         k = 3-i-j;
         if (ran[i]>=ran[j] && ran[i]>=ran[k] && ran[j]>=ran[k]) {
            hor = i;
            ver = j;
            dep = k;
            cont = 0;
         }
      }
   }

   rx  = rep[hor];
   ry  = rep[ver];
   rz  = rep[dep];
   bx  = box[hor];
   by  = box[ver];
   bz  = box[dep];
   /* SCALE DOUBLE TO UNSIGNED  */
   maxbox = bx;
   if (by>maxbox)
      maxbox = by;
   if (bz>maxbox)
      maxbox = bz;

   /*  SCALE COORDINATES TO UNSIGNED INTEGER VALUES  */
   scale = maxint / maxbox;
   iscale2 = 1.0/(scale*scale);
   ci = (unsigned *) calloc (n,sizeof(unsigned));
   cj = (unsigned *) calloc (n,sizeof(unsigned));
   ck = (unsigned *) calloc (n,sizeof(unsigned));
   ind= (unsigned *) calloc (n,sizeof(unsigned));
   if (ind==NULL) {
      printf ("   ERROR - Not enough memory to calculate neighbor list.\n");
      exit(1);
   }

   for (i=0,j=0;i<n;i++,j+=3) {
      ind[i] = i;
      ci[i] = scale * (c[j  ] - min[X]);
      cj[i] = scale * (c[j+1] - min[Y]);
      ck[i] = scale * (c[j+2] - min[Z]);
   }

   bi = scale * bx;
   bj = scale * by;
   bk = scale * bz;
   b2i = scale * (bx - rmax) - 1;
   b2j = scale * (by - rmax) - 1;
   b2k = scale * (bz - rmax) - 1;
   /*  SORT BY CURRENT X DIRECTION  */
   __sort (ci,cj,ck, ind, 0, n-1);
   /*  CONVERT CONSTANTS TO INTEGER  */
   ran1   = scale*rmax*sqrt(1) + 1;
   ran2   = scale*rmax*sqrt(2) + 1;
   ran3   = scale*rmax*sqrt(3) + 1;
   ranmin = scale*rmin         + 1;
   ranminsq = (tx=ranmin)*tx;    /*  These two variables are double  */
   ranmaxsq = (tx=ran1  )*tx;
   /*  FIND START OF ELIGILITY INTERVAL  */
   if (rx && 2*rmax<bi) {
      xm = bi + ci[0] - ran1;
      intv = n;
      while (--intv>0 && ci[intv]>xm);
      intv++;
   }
   else
      intv = 0;
   /*  FIND NEIGHBORS  */
   for (i=0;i<n;i++) {
      /*  need xj every iteration so calculation bj incrementing  */
      /*  yj and zj needed less frequently so do index            */
      if (intv==n)
         intv = 0;
      j=intv;
      while (j!=i) {
         xs = abs(ci[i]-ci[j]);
         if (rx && xs>b2i)
            xs = bi-xs;
         if (xs>=ran1)
            intv = j+1;
         else {
            ys = abs(cj[i]-cj[j]);
            if (ry && ys>b2j)
               ys = bj-ys;
            if (ys<ran1 && xs+ys<ran2) {
               zs = abs(ck[i]-ck[j]);
               if (rz && zs>b2k)
                  zs = bk-zs;
               if (zs<ran1 && xs+ys+zs<ran3) {
                  rs = (tx=xs)*tx + (ty=ys)*ty + (tz=zs)*tz;
                  if (rs>=ranminsq && rs<ranmaxsq)
                     fbnd (ind[i],ind[j],rs*iscale2);
               }
            }
         }
         if (++j==n)
            j=0;
      }
   }
   /*  RETURN  STORAGE  */
   free (ind);
   free (ck);
   free (cj);
   free (ci);
}

/*
Perform Bond calculation over coordinates pointed to by InputIndex[].
InputIndex[] contains indices of atoms stored in float **ct.
The atom indices passed to the Call Back function fbnd point to InputIndex[],
and hence range from 0 to n-1.  They DO NOT point to **ct.
*/
void bondx
(
float **ct,
unsigned *IxIndexToPart,
unsigned NumIndex,
float rmin,
float rmax,
float *Box,
unsigned *RepeatingBoundary,
void (*BondCallBack)(int,int,float)
)
   {
   float           *cx,*cy,*cz;
   unsigned        *ci,*cj,*ck;
   unsigned        *IxSortToIndex;
   unsigned        index;
   unsigned        ipart;
   unsigned        idir;
   unsigned        jdir;
   unsigned        kdir;
   unsigned        intv;
   unsigned        isort;
   unsigned        jsort;
   unsigned        hor,ver,dep, xs,ys,zs, xm;
   unsigned        ran1,ran2,ran3, bi,bj,bk, b2i,b2j,b2k, ranmin;
   char            rx,ry,rz, cont;
   float           bx,by,bz, *u,umin,umax;
   float           min[NDIR], max[NDIR], ran[NDIR];
   float           maxbox, scale, iscale2;
   float           rs,tx,ty,tz,ranminsq, ranmaxsq;
	/*  FOR 2 byte unsigned, maxint = 0x3FFFF    */
	/*      4 byte unsigned, maxint = 0x3fffffff */
	unsigned        maxint = (unsigned) ~0 >> 2;

	if ( (NumIndex==0) || ( (rmin==0.0) && (rmax==0.0) ) )
		return;

	/*   Find dimension with maximum range  */
	for (idir=X; idir<=Z; idir++)

		/*  If repeating boundary, take Box size as dimension range  */
		if (RepeatingBoundary[idir])
			{
			max[idir] = Box[idir];
			min[idir] = 0.0;
			}

		/*  .. Range is maximum span between particles */
		else
			{
			u    = ct[idir];
			umax = umin = ct[idir][0];
			for (index=0;index<NumIndex;index++)
				{
				ipart = IxIndexToPart[index];
				if (u[ipart]<umin)
					umin = u[ipart];
				else if (u[ipart]>umax)
					umax = u[ipart];
				}
			min[idir] = umin;
			max[idir] = umax;
			ran[idir] = umax-umin;
			}


	/*  ORDER DIMENSIONS BY EXTENT  */
   cont = 1;
   for (idir=X; idir<=Z && cont; idir++)
      {
      for (jdir=(idir+1)%3;jdir!=idir && cont;jdir=(jdir+1)%3)
         {
         kdir = 3-idir-jdir;
         if
         (
         ran[idir]>=ran[jdir] &&
         ran[idir]>=ran[kdir] &&
         ran[jdir]>=ran[kdir]
         )
            {
            hor = idir;
            ver = jdir;
            dep = kdir;
            cont = 0;
            }
         }
      }

   cx  = ct[hor];
   cy  = ct[ver];
   cz  = ct[dep];
   rx  = RepeatingBoundary[hor];
   ry  = RepeatingBoundary[ver];
   rz  = RepeatingBoundary[dep];
   bx  = Box[hor];
   by  = Box[ver];
   bz  = Box[dep];

   /* SCALE DOUBLE TO UNSIGNED  */
   maxbox = bx;
   if (by>maxbox)
      maxbox = by;
   if (bz>maxbox)
      maxbox = bz;

   /*  SCALE COORDINATES TO UNSIGNED INTEGER VALUES  */
   scale = maxint / maxbox;
   iscale2 = 1.0/(scale*scale);
   ci = (unsigned *) calloc (NumIndex,sizeof(unsigned));
   cj = (unsigned *) calloc (NumIndex,sizeof(unsigned));
   ck = (unsigned *) calloc (NumIndex,sizeof(unsigned));
   IxSortToIndex = (unsigned *) calloc (NumIndex,sizeof(unsigned));
   if (IxSortToIndex==NULL)
      {
      printf ("   ERROR - Not enough memory to calculate neighbor list.\n");
      exit(1);
      }

   /*
   Copy selected particles (determined by IxIndexToPart) from float array
   to scaled integer array
   */
   for (index=0;index<NumIndex;index++)
      {
      ipart      = IxIndexToPart[index];
      ci [index] = scale * (cx[ipart] - min[X]);
      cj [index] = scale * (cy[ipart] - min[Y]);
      ck [index] = scale * (cz[ipart] - min[Z]);
      }

   bi = scale * bx;
   bj = scale * by;
   bk = scale * bz;
   b2i = scale * (bx - rmax) - 1;
   b2j = scale * (by - rmax) - 1;
   b2k = scale * (bz - rmax) - 1;

   /*
   IxSortToIndex holds original particle index (as stored in input **ct)
   */
   for (index=0; index<NumIndex; index++)
      IxSortToIndex[index] = index;

   /*
   Sort atoms by current x coordinate
   When done, ci[0] will be first particle in x direction,
   cj[0],ck[0] will be y and z coordinates,
   IxSortToIndex[i] will be the original position of coordinate i in
   the ci,cj,ck arrays.
   */
   __sort (ci,cj,ck, IxSortToIndex, 0, NumIndex-1);

   /*  CONVERT CONSTANTS TO INTEGER  */
   ran1   = scale*rmax*sqrt(1) + 1;
   ran2   = scale*rmax*sqrt(2) + 1;
   ran3   = scale*rmax*sqrt(3) + 1;
   ranmin = scale*rmin         + 1;
   ranminsq = (tx=ranmin)*tx;    /*  These two variables are float  */
   ranmaxsq = (tx=ran1  )*tx;
   /*  FIND START OF ELIGILITY INTERVAL  */
   if (rx && 2*rmax<bi)
      {
      xm = bi + ci[0] - ran1;
      intv = NumIndex;
      while (--intv>0 && ci[intv]>xm);
      intv++;
      }
   else
      intv = 0;

   /*  Find NEIGHBORS  */
   for (isort=0;isort<NumIndex;isort++)
      {
      /*
      need xj every iteration so calculation bj incrementing
      yj and zj needed less frequently so do index
      */
      if (intv==NumIndex)
         intv = 0;
      jsort=intv;
      while (jsort!=isort)
         {
         xs = abs(ci[isort]-ci[jsort]);
         if (rx && xs>b2i)
            xs = bi-xs;
         if (xs>=ran1)
            intv = jsort+1;
         else
            {
            ys = abs(cj[isort]-cj[jsort]);
            if (ry && ys>b2j)
               ys = bj-ys;
            if (ys<ran1 && xs+ys<ran2)
               {
               zs = abs(ck[isort]-ck[jsort]);
               if (rz && zs>b2k)
                  zs = bk-zs;
               if (zs<ran1 && xs+ys+zs<ran3)
                  {
                  rs = (tx=xs)*tx + (ty=ys)*ty + (tz=zs)*tz;
                  if (rs>=ranminsq && rs<ranmaxsq)
                     BondCallBack
                        (
                        IxSortToIndex[isort],
                        IxSortToIndex[jsort],
                        rs*iscale2
                        );
                  }
               }
            }
         if (++jsort==NumIndex)
            jsort=0;
         }
      }

   /*  RETURN  STORAGE  */
   free (IxSortToIndex);
   free (ck);
   free (cj);
   free (ci);
   }


void CalcRDF
(
double *c, unsigned n, double rmin, double rmax,
double *box, unsigned *rep, long *table, int ntable
)
{
   int i;

   /*  TEST RADIUS LIMITS  */
   if (rmin>=rmax)
      return;
   /*  SCALING FACTOR  */
   gl_rmin   = rmin;
   gl_rsqmin = rmin*rmin;
   gl_rsqmax = rmax*rmax;
   gl_table  = table;
   gl_fac    = ntable / (rmax-rmin);
   /*  INITIALIZE TABLE  */
   for (i=0;i<ntable;i++)
      table[i]=0;

   /*  CALCULATE RDF  */
   bondd (c, n, rmin, rmax, box, rep, RDF_func);
}


void CalcTypeRDF
(
double *Coord,
UCHAR  *Type,
unsigned NumCoord,
double MinRadius,
double MaxRadius,
int    FirstType,
int    SecondType,
double *Box,
unsigned *Repeat,
long     *Table,
int      NumTable
)
   {
   int Index;

   /*  TEST RADIUS LIMITS  */
   if (MinRadius>=MaxRadius)
      return;

   /*  Set Global Variables for Callback Function  */
   gl_rmin   = MinRadius;
   gl_rsqmin = MinRadius*MinRadius;
   gl_rsqmax = MaxRadius*MaxRadius;
   gl_table  = Table;
   gl_fac    = NumTable / (MaxRadius-MinRadius);
   gl_TypeList = Type;
   gl_TypePattern = GetTypePattern (FirstType, SecondType);

   /*  INITIALIZE TABLE  */
   for (Index=0; Index<NumTable; Index++)
      Table[Index] = 0;

   /*  CALCULATE RDF  */
   bondd
      (
      Coord,
      NumCoord,
      MinRadius,
      MaxRadius,
      Box,
      Repeat,
      TypeRDF_func
      );
}

/*
************************************************************************
Local Functions
************************************************************************
*/

void __sort ( unsigned *x,
            unsigned *y,
            unsigned *z,
            unsigned *ind,
            int lo,
            int hi  )
{  int i,j; unsigned t,u;
   i=lo;
   j=hi;
   t=x[(lo+hi)/2];
   do {
      while (x[i]<t)
         i++;
      while (t<x[j])
         j--;
      if (i<=j)  {
         if (i<j) {
            u =   x[i];    x[i] =   x[j];    x[j] = u;
            u =   y[i];    y[i] =   y[j];    y[j] = u;
            u =   z[i];    z[i] =   z[j];    z[j] = u;
            u = ind[i];  ind[i] = ind[j];  ind[j] = u;
         }
         i++;
         j--;
       }
   } while (i<=j);
   if (lo<j ) __sort (x,y,z, ind, lo, j );
   if (i <hi) __sort (x,y,z, ind, i , hi);
}

#pragma argsused
void RDF_func (int i, int j, double rsq)
{
   double r;
   int    index;

   if (rsq>gl_rsqmin && rsq<gl_rsqmax)
      {
      r = sqrt(rsq);
      index = gl_fac * (r-gl_rmin);
      gl_table[index]++;
      }
}

void TypeRDF_func (int i, int j, double rsq)
   {
   double r;
   int index;

   /*  Test type value against the value wanted  */
   if ( gl_TypePattern == GetTypePattern(gl_TypeList[i], gl_TypeList[j]) )
      {
      if (rsq>gl_rsqmin && rsq<gl_rsqmax)
         {
         r = sqrt(rsq);
         index = gl_fac * (r-gl_rmin);
         gl_table[index]++;
         }
      }
   }

int GetTypePattern (int FirstType, int SecondType)
   {
   if (FirstType<SecondType)
      return  ((FirstType  & 0x00ff) << 8) | SecondType;
   else
      return  ((SecondType & 0x00ff) << 8) | FirstType;
   }
