/*

SOURCE CODE FOR READING AND WRITING RCV FILES
Uses stepdata structure, modelled after Pascal version of PLOTL

READRCV : read a single step from a rcv file.
WRITERCV: write a single step to a rcv file.

 6 Jun 1991
28 Oct 1991
 6 Nov 1991 - add code to leave types alone if selkeep flag on
15 Nov 1991 - set trans to 0 if boundrep is on
17 Nov 1991 - streamline file read
21 Nov 1991 - error with trans, was setting to zero when boundrep data
              was present rather than vice versa.
15 May 1992 - error occured in WRITERCV with line maxtbl += 1.5;.  Computer
              would hang when compiled with CLIPRCV.  Changed to maxtbl += 1;.

?? Dec 1995 - Added COR routines
23 Feb 1996 - Write elements of FileHeader struct individually
              instead of all together.  This eliminates porting
              problems on machines on RS6000 which pad the
              structure with bytes between variables, becuase
              writing the entire structure with one fwrite()
              command writes the padding bytes as well as the
              variables, which will confused programs run on
              other machines which try to read the resulting file.

*/

/*
************************************************************************
Compile Switches
************************************************************************
*/

/*
************************************************************************
File Includes
************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsub.h>
#include <time.h>
#include <cor0.h>



/*
************************************************************************
Defines
************************************************************************
*/

/*  Define default float type  */
#define FLOAT float

/*  Maximum value of word  */
#define MAX_WORD 0xffff

/*  Default string length  */
#define STRING_LENGTH 256

/*  Types of boundaries  */
#define BD_ORTHONORMAL  1     /*  Standard box boundary         */
#define BD_MONOCLINIC   2     /*  Parallel-piped boundary       */
#define BC_ARC          3     /*  Curved "race track" boundary  */

/*  Number of characters in file type  */
#define LEN_FILE_TYPE 3

/*
Define types for standardizing file reads between machines
  - Following definitions valid for PC, Linux and AIX
*/
#define INT2  short
#define INT4  int
#define WORD1 unsigned char
#define WORD2 unsigned short
#define REAL4 float
#define REAL8 double

/*
************************************************************************
Macros
************************************************************************
*/

/*  Macro for reversing byte order  */
#define REVERSE(VAR) \
   ReverseVar (&(VAR), sizeof(VAR));

/*  Macro for string write  */
#define FWRITESTR(STRING,FILE) \
   fwrite (STRING, strlen(STRING), (1), (FILE));

/*  Macro for block read  */
#define FREAD(VAR,FILE,N) \
   fread_return_m = fread (&(VAR), sizeof(VAR), (N), (FILE));

#define FGETS(VAR, N, FILE) \
   fgets_return_m = fgets ((VAR), (N), (FILE));

#define FSCANF(FILE, FMT, VAR) \
   fscanf_return_m = fscanf((FILE), (FMT), (VAR))

/*  Macro for block write  */
#define FWRITE(VAR,FILE,N) \
   fwrite (&(VAR), sizeof(VAR), (N), (FILE));

/*  Macro to reallocate array  */
#define REALLOCATE(ARRAY,TYPE,NUM) \
   if ((ARRAY)==NULL) \
      ARRAY = (TYPE *) calloc ((NUM), sizeof(TYPE)); \
   else \
      ARRAY = realloc ((ARRAY), (NUM)*sizeof(TYPE) );

/*  Macro to free array  */
#define FREE(ARRAY) \
   if ((ARRAY)!=NULL) \
      { \
      free (ARRAY); \
      (ARRAY) = NULL; \
      }

/*  Macro for loops  */
#define LOOP(INDEX,LIMIT) \
   for (INDEX=0; INDEX<(LIMIT); INDEX++)



#define WRITEMSG \
	printf ("In File %s at Line %i: ", __FILE__, __LINE__); 

#define WRITEVAR(VAL,FORMAT) \
	printf ("File %s Line %i: ", __FILE__, __LINE__); \
   printf ("%s = ", #VAL); \
   printf (#FORMAT, VAL); \
   printf ("\n");

/*
************************************************************************
Type Definitions
************************************************************************
*/
typedef struct
   {
   INT2   EndianWord;
   INT4   Run;
   INT4   Step;
   REAL8  Time;
   WORD2  BoundaryType;
   WORD1  Surf[NDIR];
   REAL8  Box [NDIR];
   REAL8  Etot;
   REAL8  Epot;
   REAL8  Ekin;
   REAL8  Ebath;
   REAL8  Min [NDIR];
   REAL8  Max [NDIR];
   INT4   Np;
   }
   CorFileHeader_t;


/*
************************************************************************
Module Wide Variables
************************************************************************
*/

/* TRANSLATE: CONVERTS TWO Code byte VALUES INTO integer FROM 0 TO 9999 */
/* THIS TABLE TRANSLATES BETWEEN charACTERS SENT FROM PURDUE
 CDC 205 AND THE CRAY AT LLL TO THE  IBM PC.  THE SEND PROGRAM
 TRANSLATES THE FOLLOWING charACTERS
 CHARACTER      REASON
 ---------      -------
 27 -> 17       CRAY
 28 -> 14
 30 -> 15
 31 -> 25       CYBER LINE FEED CONTROL CHARACTER
 92 -> 16
(94 -> 20)      ASCII TO EBCDIC CONVERSION  )
(124-> 21)      ASCII TO EBCDIC CONVERSION  )
 126-> 20       VAX MAIL UTILITY
*/

int off  =   27;
int base =  100; /* NOTE:  Base is fixed in values of table2  */
int off2 = 2727; /* off2 = off * (base + 1)  */
int table[128] =
   {  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,
          11,  12,  13,  28,  30,  92,  27,  18,  19, 126,
          21,  22,  23,  24,  31,  26,  27,  28,  29,  30,
          31,  32,  33,  34,  35,  36,  37,  38,  39,  40,
          41,  42,  43,  44,  45,  46,  47,  48,  49,  50,
          51,  52,  53,  54,  55,  56,  57,  58,  59,  60,
          61,  62,  63,  64,  65,  66,  67,  68,  69,  70,
          71,  72,  73,  74,  75,  76,  77,  78,  79,  80,
          81,  82,  83,  84,  85,  86,  87,  88,  89,  90,
          91,  92,  93,  94,  95,  96,  97,  98,  99, 100,
         101, 102, 103, 104, 105, 106, 107, 108, 109, 110,
         111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
         121, 122, 123, 124, 125, 126, 127                };

int table2[128] =
 {  000,   100,   200,   300,   400,   500,   600,   700,   800,   900,  1000,
          1100,  1200,  1300,  2800,  3000,  9200,  2700,  1800,  1900, 12600,
          2100,  2200,  2300,  2400,  3100,  2600,  2700,  2800,  2900,  3000,
          3100,  3200,  3300,  3400,  3500,  3600,  3700,  3800,  3900,  4000,
          4100,  4200,  4300,  4400,  4500,  4600,  4700,  4800,  4900,  5000,
          5100,  5200,  5300,  5400,  5500,  5600,  5700,  5800,  5900,  6000,
          6100,  6200,  6300,  6400,  6500,  6600,  6700,  6800,  6900,  7000,
          7100,  7200,  7300,  7400,  7500,  7600,  7700,  7800,  7900,  8000,
          8100,  8200,  8300,  8400,  8500,  8600,  8700,  8800,  8900,  9000,
          9100,  9200,  9300,  9400,  9500,  9600,  9700,  9800,  9900, 10000,
         10100, 10200, 10300, 10400, 10500, 10600, 10700, 10800, 10900, 11000,
         11100, 11200, 11300, 11400, 11500, 11600, 11700, 11800, 11900, 12000,
         12100, 12200, 12300, 12400, 12500, 12600, 12700              };

INT2  EndianWord_m = 0x0001;
INT2  SwitchedEndianWord_m = 0x0100;
INT4  BoundaryType_m = BD_ORTHONORMAL;

char *VersionStr_m = "COR FILE: Version 1.0.0";
char *CompareStr_m = "COR FILE";

/*  Used to get catch values returned from file reading functions, and
    thereby avoid compiler warnings */
size_t fread_return_m;
char  *fgets_return_m;
int    fscanf_return_m;


/*
************************************************************************
Local Function Prototypes
************************************************************************
*/
void GetNextStringFromFile (char *InputString, int Length, FILE *InputFile);
static char *GetCurrentTimeStr (void);
int GetCoordFileType (char *FileName);
char *GetFileType (char *FileName);
void ReverseVar (void *WordPtr, int SizeVar);
void ReverseHeader (CorFileHeader_t *);

void ScaleCoord (double, stepdata *);


/*
************************************************************************
Exported Routines
************************************************************************
*/


int readrcv (FILE *fin, stepdata *cur)
   {
   int res  = 10000;
   float     scale[3], tablefactor;
   unsigned  np,i, n,iline, ntable;
   long      maxtable;
   int       teof,xlo,xhi,ylo,yhi,zlo,zhi;
   int done;
   char instr[NSTR], *tstr;

   /* READ FILE HEADER                       */
   xlo = getc (fin); /*  GOBBLE UP FIRST CHARACTER  */
   if (xlo==EOF)
      return EOF;
   FGETS(instr, NSTR, fin);
   teof = sscanf (instr, "%li %li %f %f %f %i %i\n",
      &cur->run, &cur->step, &cur->time, &cur->tempc, &cur->temp,
      &np, &cur->ndim);
   if (teof==EOF)
      return EOF;
   /*  AJUST NUMBER OF POINTS  */
   np--;
   for (i=0;i<NTITLE;i++)
      {
      FGETS (cur->title[i], NSTR, fin);
      strtok(cur->title[i], "\n");
      }

   /*  READ MISC DATA LINE  */
   FGETS  (instr, NSTR, fin);
   tstr = instr;
   cur->mass   = dblstr (&tstr);
   cur->vibp   = dblstr (&tstr);
   cur->dtime  = dblstr (&tstr);
   cur->cutoff = dblstr (&tstr);

   /*  READ Strain AND Stress  */
   if (*tstr!=0)
      {
      cur->strain = dblstr (&tstr);
      cur->stress = dblstr (&tstr);
      }
   else
      cur->stress = cur->strain = 0.0;

   /*  READ BOX  */
   FGETS  (instr, NSTR, fin);
   tstr = instr;
   for (i=0;i<3;i++)
      cur->box[i] = dblstr(&tstr);

   /*  READ Trans IF PRESENT  */
   if (*tstr!=0)
      for (i=0;i<3;i++)
         cur->trans[i] = dblstr(&tstr);
   else
      for (i=0;i<3;i++)
          cur->trans[i] = 0;

   /*  BOUNDARY IF PRESENT  */
   if (*tstr!=0)
      {
      for (i=0;i<3;i++)
         cur->boundrep[i] = intstr(&tstr);
      }
   else
      for (i=0;i<3;i++)
         {
         cur->boundrep[i] = 0;
         cur->trans[i] = 0.0;
         }

   for (i=0;i<3;i++)
      scale[i] = cur->box[i] / res;

   /*  Allocate Select Info  */
   if (cur->selkeep)
      {
      REALLOCATE (cur->sel, BYTE, np)
      REALLOCATE (cur->tag, BYTE, np)
      }
   else
      {
      FREE (cur->sel)
      FREE (cur->tag)
      }

   /*  Allocate Space  */
   REALLOCATE (cur->t,    BYTE,  np)
   REALLOCATE (cur->c[X], float, np)
   REALLOCATE (cur->c[Y], float, np)
   REALLOCATE (cur->c[Z], float, np)
   if (cur->c[Z]==NULL)
      {
      fprintf (stderr,
         "ERROR (readrcv):  No room to allocate %i new (%i total) particles.\n",
         np-cur->np, np);
      exit(1);
      }
   cur->np = np;

   /*  Read COORDINATES IN CodeD FORM   */
   n = 0;
   iline = 0;
   do
      /* UNTIL END-OF-FILE  OR  UNTIL ASCII 0 Code FOUND */
      {
      i = 0;
      FGETS (instr, NSTR, fin);
      tstr = instr;
      do
         /* UNTIL 13 POSITIONS Read IN   OR  END-OF-FILE  OR  0 FOUND*/
         {
         xlo = *(tstr++);
         if (xlo!='\n' && xlo!=0)
            {
            xhi = *(tstr++);
            ylo = *(tstr++);
            yhi = *(tstr++);
            zlo = *(tstr++);
            zhi = *(tstr++);
            if (feof(fin))
               {
               fprintf (stderr, "ERROR (readrcv):  Premature end of file in step %li.\n", cur->step);
               exit(1);
               }
            /*  TEST FOR N TOO BIG  */
            if (n==np)
               {
               fprintf (stderr, "ERROR (readrcv): Data is corrupted.\n");
               fprintf (stderr, "                 Number of particles in header (%i) and body (%i)\n", np, n);
               fprintf (stderr, "                 for run (%li) step (%li) do not agree.\n", cur->run, cur->step);
               exit(1);
               }
            cur->c[0][n] = scale[0] * (table[xlo]+table2[xhi]-off2) - cur->trans[0];
            cur->c[1][n] = scale[1] * (table[ylo]+table2[yhi]-off2) - cur->trans[1];
            cur->c[2][n] = scale[2] * (table[zlo]+table2[zhi]-off2) - cur->trans[2];
            if (!cur->selkeep)
               cur->t[n] = 1;
            n++;
            i++;
            }
         } while (i!=13 && xlo !=0 && xlo!='\n' && n!=np);
      iline++;
      } while (xlo !=0 && xlo!='\n' && n!=np);

   if (n<np)
      {
      fprintf (stderr, "ERROR (readrcv): Only %i of %i points read for step %li.\n",
                        n, np, cur->step);
      exit(1);
      }


   /*  Read RDF table  */
   FSCANF (fin, "%li\n", &maxtable);
   tablefactor = maxtable/10000.0;
   ntable = 0;
   do
      {
      /* UNTIL END-OF-FILE  OR  UNTIL ASCII 0 Code FOUND */
      iline = 0;
      do
       /*     UNTIL 13 POSITIONS Read IN
                OR  END-OF-FILE  OR  0 FOUND    */
         {
         xlo = getc(fin);
         done = (xlo==13 || xlo==0 || xlo==EOF);
         if (!done)
            {
            xhi = getc(fin);
            iline++;
            if (xhi!=0 && xhi!=13 && xhi!=EOF)
               cur->rdftable[ntable++] =
                       tablefactor * (table[xlo]+table2[xhi]-off2);
            }
         done = (done || ntable==100);
         } while (iline<39 && !done);

      /*   Goto end of line  */
      while (getc(fin)!='\n');

      } while (!done);


   /*  Find min and maximum coordinate  */
   CalcMinMaxCoord (cur, FALSE, FALSE);


   /*  EVERYTHING OK:  RETURN not eof  */
   return 0;
   }
/*  END READRCV  */




void writercv (FILE *fout, stepdata *s, int selfl)
   {
   long maxtbl;
   int i,j,k, nchar, code;
   int line[79];
   int base = 100;
   int off  =  27;
   int res  = 10000;
   int nbin = 100;
   float boxsc[3];
   float tr[3];
   int   bx[3];

   /*
   Find min and max coordinate
    - use only selected particles if selfl
    - use repeating boundary conditions
   */
   CalcMinMaxCoord (s, selfl, TRUE);

   /*  ADJUST BOX IF FREE SURFACE      */
   /*  ADJUST PARTICLE TRANSLATION     */
   /*  SET BOUNDARY FLAG (FOR OUTPUT)  */

   for (i=0;i<3;i++) {
      if (!s->boundrep[i])
         {
         boxsc[i] = ( s->MaxCoord[i] - s->MinCoord[i] ) * 1.01;
         tr   [i] = -s->MinCoord[i];
         bx   [i] = 0;
         }
      else {
         boxsc[i] = s->box[i];
         tr   [i] = 0;
         bx   [i] = 1;
      }
   }

   /*  PLACE REPEATING PARTICLES INSIDE StepData  */
   for (i=0;i<3;i++)
      if (s->boundrep[i])
         for (j=0;j<s->np;j++)
            if (s->c[i][j] > s->box[i])
               s->c[i][j] -= s->box[i];
            else if (s->c[i][j]<0)
               s->c[i][j] += s->box[i];


   /*  SEND HEADER DATA  */
   if (selfl)
      fprintf (fout, "  %li %li %f %f %f %i 3\n", s->run, s->step, s->time,
                     s->tempc, s->temp, s->nsel+1);
   else
      fprintf (fout, "  %li %li %f %f %f %i 3\n", s->run, s->step, s->time,
                     s->tempc, s->temp, s->np+1);

   for (i=0;i<NTITLE;i++)
      if (s->title[i]==NULL)
         fprintf (fout, "* title *\n");
      else
         fprintf (fout, "%s\n", s->title[i]);

   fprintf (fout, "%14e  %15e  %10.4f  %10.4f %13e %13e\n",
            s->mass, s->vibp, s->dtime, s->cutoff, s->strain, s->stress);
   fprintf (fout, "%10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %2i %2i %2i\n",
            boxsc[0], boxsc[1], boxsc[2],
            tr[0], tr[1], tr[2], bx[0], bx[1], bx[2]);



   /*  SEND CODED PARTICLE COORDINTES  */
   nchar = 0;
   for (i=0;i<s->np;i++) {
      if (!selfl || s->sel==NULL || s->sel[i]) {
         for (j=0;j<3;j++) {
            code = res * (s->c[j][i] + tr[j]) / boxsc[j];
            line[nchar++] = code % base + off;
            line[nchar++] = code / base + off;
         }
         if (nchar==78) {
            line[nchar++] = '.';
            for (k=0;k<nchar;k++)
               putc (line[k], fout);
            putc ('\n', fout);
            nchar = 0;
          }
          else if (nchar>78) {
             fprintf (stderr,
                "INTERNAL ERROR (File %s Line %i): nchar (%i) > 78.\n", 
						__FILE__, __LINE__, nchar);
             exit(1);
          }


      }
   }

   /*  SEND REMAINING CHARACTERS  */
   if (nchar>0) {
      for (k=0;k<nchar;k++)
         putc (line[k], fout);
      putc ('\n', fout);
   }


   /*  RDF TABLE  */
   nbin  = 100;
   maxtbl = s->rdftable[0];
   for (i=1;i<nbin;i++)
      if (maxtbl<s->rdftable[i])
         maxtbl = s->rdftable[i];
   /* adjust maxtbl to avoid overflow  */
   maxtbl += 1;
   if (maxtbl==0)
      maxtbl = 1;
   fprintf (fout, "%li\n", maxtbl);
   nchar=0;
   for (i=0;i<nbin;i++) {
      code = res * s->rdftable[i] / maxtbl + 0.5;
      line[nchar++] = code % base + off;
      line[nchar++] = code / base + off;
      if (nchar==78) {
         line[nchar++] = 0;
         for (k=0;k<nchar;k++)
            putc (line[k], fout);
         putc ('\n', fout);
         nchar = 0;
      }
      else if (nchar>78) {
         fprintf (stderr,
            "INTERNAL ERROR 2 (File %s Line %i): nchar (%i) > 78.\n", 
					__FILE__, __LINE__, nchar);
         exit(1);
      }
   }

   /*  SEND REMAINING CHARACTERS  */
   if (nchar>0)  {
      for (k=0;k<nchar;k++)
         putc (line[k], fout);
      putc ('\n', fout);
   }

}
/*  END WRITERCV  */




/*  Free Step Data  */
void FreeStep (stepdata *cur)
   {
   /*  Free old Variables  */
   if (cur->c[X]!=NULL)
      free (cur->c[X]);
   if (cur->c[Y]!=NULL)
      free (cur->c[Y]);
   if (cur->c[Z]!=NULL)
      free (cur->c[Z]);
   if (cur->t    !=NULL)
      free (cur->t    );
   if (cur->cd[X]!=NULL)
      free (cur->cd[X]);
   if (cur->cd[Y]!=NULL)
      free (cur->cd[Y]);
   if (cur->cd[Z]!=NULL)
      free (cur->cd[Z]);
   if (cur->radius!=NULL)
      free (cur->radius);
   if (cur->color!=NULL)
      free (cur->color );
   if (cur->symbol!=NULL)
      free (cur->symbol);
   if (cur->fill!=NULL)
      free (cur->fill  );
   if (cur->dispfl!=NULL)
      free (cur->dispfl);
   }




/*  Allocate Step Data  (not coordinates and related arrays)  */
stepdata *make_step ()
   {
   return (stepdata *) calloc ((unsigned) 1, sizeof(stepdata) );
   }




void readtype (FILE *fin, stepdata *cur)
   {
   int i,n,nalloc;
   int InputType;
   FSCANF (fin, "TYPELIST%i", &n);
   if (n>cur->np)
      {
      nalloc  = n;
      cur->np = n;
      }
   else
      {
      nalloc = cur->np;
      n      = cur->np;
      }

   /*  RE ALLOCATE STORAGE  */
   if (cur->t!=NULL)
      cur->t = (unsigned char *)
                  realloc (cur->t, nalloc*sizeof(unsigned char));

   /*  ALLOCATE NEW STORAGE  */
   else
      cur->t = (unsigned char *) calloc (nalloc, sizeof(unsigned char));

   /*  CHECK ALLOCATION  */
   if (cur->t==NULL)
      {
      fprintf (stderr, "ERROR in readtype:  Insufficient memory for %i types.\n",
         nalloc);
      exit(1);
      }

   /*  READ TYPES  */
   i = 0;
   while (EOF!=fscanf(fin,"%i", &InputType)  &&  i<n)
      {
      cur->t[i] = InputType-1;
      i++;
      }

   /*  Test for premature end of file  */
   if (i<n)
      {
      fprintf (stderr, "ERROR in readtype:  premature end of file.\n");
      exit(1);
      }
   }




void readilist(FILE *fin, stepdata *cur)
   {
   unsigned i,n,is,nalloc,*tind;
   FSCANF (fin, "ILIST%i", &n);
   /*  ALLOCATE TEMPORARY INDEX STORAGE  */
   tind = (unsigned *)  calloc (n, sizeof(unsigned));
   if (tind==NULL)
      {
      fprintf (stderr,
      "ERROR in readilist:  Insufficient memory for allocating temporary indices.\n");
      exit(1);
      }

   /*  READ ILIST  */
   i = 0;
   while (EOF!=fscanf(fin,"%i", &is)  &&  i<n)
      {
      tind[i] = is;
      i++;
      }
   if (i<n)
      {
      fprintf (stderr, "ERROR in readilist:  premature end of file.\n");
      exit(1);
      }


   /*  FIND MAXIMUM INDEX  */
   nalloc = tind[0];
   for (i=1;i<n;i++)
      if (tind[i]>nalloc)
         nalloc = tind[i];


   /*  ADJUST STORAGE FOR SELECT FLAGS  */
   if (cur->np>nalloc)
      nalloc = cur->np;
   if (cur->sel==NULL)
      cur->sel = (unsigned char *) calloc (nalloc, sizeof(unsigned));
   else
      cur->sel = (unsigned char *) realloc(cur->sel, nalloc*sizeof(unsigned));
   if (cur->sel==NULL) {
      fprintf (stderr,
      "ERROR in readilist:  Insufficient memory for allocating select flags.\n");
      exit(1);
   }


   /*  SET SELECT FLAGS */
   for (i=0;i<n;i++)
      cur->sel[tind[i]-1] = 1;
   cur->nsel = n;


   /*  FREE STORAGE  */
   free(tind);
   }





/*
************************************************************************
Cor Routines
************************************************************************
*/

void WriteCorFile (FILE *OutputFile, stepdata *a, BOOLEAN UseSelect)
   {
   CorFileHeader_t FileHeader;
   int    ititle;
   int    ipart;
   int    idir;
   FLOAT  Scale[NDIR];
   char   OutputStr[STRING_LENGTH];
   char  *LineFeed = "\n\r";
   INT4   OutputNP;
   WORD2  WordCoord[NDIR];


   /*  Check for no selected particles  */
   if ( (UseSelect && a->nsel==0) || a->np==0)
      return;

   /*
   Calc min and max coordinates
    - heed the Select flag
    - do not use boundary conditions
   */
   CalcMinMaxCoord (a, UseSelect, FALSE);


   /*  Detemine number of output particles  */
   if (UseSelect)
      OutputNP = a->nsel;
   else
      OutputNP = a->np;


   /*
   **********************************
   Write Text portions of File Header
   **********************************
   */

   /*  Write file type and version  */
   FWRITESTR (VersionStr_m, OutputFile)
   FWRITESTR (LineFeed, OutputFile)

   FWRITESTR ("DATE: ", OutputFile)
   FWRITESTR (GetCurrentTimeStr(), OutputFile)
   FWRITESTR (LineFeed, OutputFile)

   /*  Write run and step  */
   sprintf   (OutputStr, "RUN: %li", a->run);
   FWRITESTR (OutputStr, OutputFile)
   FWRITESTR (LineFeed,  OutputFile)

   sprintf   (OutputStr, "STEP: %li", a->step);
   FWRITESTR (OutputStr, OutputFile)
   FWRITESTR (LineFeed,  OutputFile)

   FWRITESTR ("TITLE:", OutputFile)
   FWRITESTR (LineFeed,  OutputFile)


   /*  Movie file comments go here, at header end  */
   LOOP (ititle, 8)
      {
      FWRITESTR (a->title[ititle], OutputFile)
      FWRITESTR (LineFeed, OutputFile)
      }

   /*  Write header end (DOS End-of-file character)  */
   FWRITESTR ("\032", OutputFile)

   /*  Set values of file header  */
   FileHeader.EndianWord   = EndianWord_m;
   FileHeader.Run          = a->run;
   FileHeader.Step         = a->step;
   FileHeader.Time         = a->time;
   FileHeader.BoundaryType = BoundaryType_m;
   FileHeader.Surf[X]      = !a->boundrep[X];
   FileHeader.Surf[Y]      = !a->boundrep[Y];
   FileHeader.Surf[Z]      = !a->boundrep[Z];
   FileHeader.Box[X]       = a->box[X];
   FileHeader.Box[Y]       = a->box[Y];
   FileHeader.Box[Z]       = a->box[Z];
   FileHeader.Etot         = a->etot ;
   FileHeader.Epot         = a->epot ;
   FileHeader.Ekin         = a->ekin ;
   FileHeader.Ebath        = a->ebath;
   FileHeader.Min[X]       = a->MinCoord[X];
   FileHeader.Min[Y]       = a->MinCoord[Y];
   FileHeader.Min[Z]       = a->MinCoord[Z];
   FileHeader.Max[X]       = a->MaxCoord[X];
   FileHeader.Max[Y]       = a->MaxCoord[Y];
   FileHeader.Max[Z]       = a->MaxCoord[Z];
   FileHeader.Np           = OutputNP;

   /*  Write output header  */
   FWRITE (FileHeader.EndianWord,    OutputFile, 1)
   FWRITE (FileHeader.Run,           OutputFile, 1)
   FWRITE (FileHeader.Step,          OutputFile, 1)
   FWRITE (FileHeader.Time,          OutputFile, 1)
   FWRITE (FileHeader.BoundaryType,  OutputFile, 1)
   FWRITE (FileHeader.Surf[X],       OutputFile, 1)
   FWRITE (FileHeader.Surf[Y],       OutputFile, 1)
   FWRITE (FileHeader.Surf[Z],       OutputFile, 1)
   FWRITE (FileHeader.Box[X],        OutputFile, 1)
   FWRITE (FileHeader.Box[Y],        OutputFile, 1)
   FWRITE (FileHeader.Box[Z],        OutputFile, 1)
   FWRITE (FileHeader.Etot,          OutputFile, 1)
   FWRITE (FileHeader.Epot,          OutputFile, 1)
   FWRITE (FileHeader.Ekin,          OutputFile, 1)
   FWRITE (FileHeader.Ebath,         OutputFile, 1)
   FWRITE (FileHeader.Min[X],        OutputFile, 1)
   FWRITE (FileHeader.Min[Y],        OutputFile, 1)
   FWRITE (FileHeader.Min[Z],        OutputFile, 1)
   FWRITE (FileHeader.Max[X],        OutputFile, 1)
   FWRITE (FileHeader.Max[Y],        OutputFile, 1)
   FWRITE (FileHeader.Max[Z],        OutputFile, 1)
   FWRITE (FileHeader.Np,            OutputFile, 1)

#if 0
   /*  Write file header  */
   fwrite   (&FileHeader, sizeof(FileHeader), 1, OutputFile);
#endif


   /*  Write particle types  */
   LOOP (ipart, a->np)
      if (!UseSelect || a->sel[ipart])
         fwrite (&a->t[ipart], sizeof(WORD1), 1, OutputFile);

   /*  Find scale for particles  */
   LOOP (idir, NDIR)
      {
      if (a->MaxCoord[idir]==a->MinCoord[idir])
         Scale[idir] = 0.0;
      else
         Scale[idir] =
            MAX_WORD / (a->MaxCoord[idir] - a->MinCoord[idir]);
      }

   /*  Write particle coordinates  */
   LOOP (ipart, a->np)
      if (!UseSelect || a->sel[ipart])
         {
         LOOP (idir, NDIR)
            WordCoord[idir] =
               Scale[idir] * (a->c[idir][ipart] - a->MinCoord[idir]);
         fwrite (WordCoord, sizeof(WORD2), NDIR, OutputFile);
         }


   }
/*   ... WriteCorFile  */


int ReadCorFile (FILE *InputFile, stepdata *a)
   {
   int    ititle;
   int    ipart;
   int    idir;
   int    InputNP;
   double Scale[NDIR];
   char   InputChar;
   char   InputStr[STRING_LENGTH];
   CorFileHeader_t FileHeader;
   BOOLEAN IsReverse;
   WORD2   WordCoord[NDIR];
   double  Min[NDIR];
   double  Max[NDIR];

   /*  Determine if input file is a cor file  */
   GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);

   /*  Test for end-of-file  */
   if (feof(InputFile))
      return EOF;

   if (strncmp(InputStr, CompareStr_m, strlen(CompareStr_m))!=0)
      {
      printf ("ERROR READING COR FILE: \n");
      printf ("  Input file is not a COR file.\n");
      return EOF;
      }

   /*  Read input file version  */
   if (strcmp(InputStr, VersionStr_m)!=0)
      {
      printf ("ERROR READING COR FILE: \n");
      printf ("  Input file is version: %s\n", InputStr);
      printf ("  Can only read version: %s\n", VersionStr_m);
      return EOF;
      }

   /*  Skip date, run, step, title, strings  */
   GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);
   GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);
   GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);
   GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);


   /*  Read title lines  */
   LOOP (ititle, 8)
      {
      GetNextStringFromFile (InputStr, STRING_LENGTH, InputFile);
      strcpy (a->title[ititle], InputStr);
      }


   /*  Next character should be DOS End-of-file character  */
   FREAD(InputChar, InputFile, 1)
   if (InputChar != '\032')
      {
      printf ("ERROR:  COR file is corrupted.\n");
      return EOF;
      }

   FREAD (FileHeader.EndianWord,    InputFile, 1)
   FREAD (FileHeader.Run,           InputFile, 1)
   FREAD (FileHeader.Step,          InputFile, 1)
   FREAD (FileHeader.Time,          InputFile, 1)
   FREAD (FileHeader.BoundaryType,  InputFile, 1)
   FREAD (FileHeader.Surf[X],       InputFile, 1)
   FREAD (FileHeader.Surf[Y],       InputFile, 1)
   FREAD (FileHeader.Surf[Z],       InputFile, 1)
   FREAD (FileHeader.Box[X],        InputFile, 1)
   FREAD (FileHeader.Box[Y],        InputFile, 1)
   FREAD (FileHeader.Box[Z],        InputFile, 1)
   FREAD (FileHeader.Etot,          InputFile, 1)
   FREAD (FileHeader.Epot,          InputFile, 1)
   FREAD (FileHeader.Ekin,          InputFile, 1)
   FREAD (FileHeader.Ebath,         InputFile, 1)
   FREAD (FileHeader.Min[X],        InputFile, 1)
   FREAD (FileHeader.Min[Y],        InputFile, 1)
   FREAD (FileHeader.Min[Z],        InputFile, 1)
   FREAD (FileHeader.Max[X],        InputFile, 1)
   FREAD (FileHeader.Max[Y],        InputFile, 1)
   FREAD (FileHeader.Max[Z],        InputFile, 1)
   FREAD (FileHeader.Np,            InputFile, 1)


   /*  Determine if data from different endian machine  */
   IsReverse = (FileHeader.EndianWord != EndianWord_m);

   /*  Reverse variables if necessary  */
   if (IsReverse)
      ReverseHeader (&FileHeader);

   /*  Copy header to stepdata  */
   a->run           =  FileHeader.Run;
   a->step          =  FileHeader.Step;
   a->time          =  FileHeader.Time;
   a->boundrep[X]   = !FileHeader.Surf[X];
   a->boundrep[Y]   = !FileHeader.Surf[Y];
   a->boundrep[Z]   = !FileHeader.Surf[Z];
   a->box[X]        =  FileHeader.Box[X];
   a->box[Y]        =  FileHeader.Box[Y];
   a->box[Z]        =  FileHeader.Box[Z];
   a->etot          =  FileHeader.Etot;
   a->epot          =  FileHeader.Epot;
   a->ekin          =  FileHeader.Ekin;
   a->ebath         =  FileHeader.Ebath;
   Min[X]           =  FileHeader.Min[X];
   Min[Y]           =  FileHeader.Min[Y];
   Min[Z]           =  FileHeader.Min[Z];
   Max[X]           =  FileHeader.Max[X];
   Max[Y]           =  FileHeader.Max[Y];
   Max[Z]           =  FileHeader.Max[Z];
   InputNP          =  (int) FileHeader.Np;

   /*  Allocate space for types and particles  */
   if (a->selkeep)
      {
      REALLOCATE (a->sel, BYTE, InputNP)
      REALLOCATE (a->tag, BYTE, InputNP)
      }
   else
      {
      FREE (a->sel)
      FREE (a->tag)
      }

   /*  Allocate Space  */
   REALLOCATE (a->t,    BYTE,  InputNP)
   REALLOCATE (a->c[X], float, InputNP)
   REALLOCATE (a->c[Y], float, InputNP)
   REALLOCATE (a->c[Z], float, InputNP)
   if (a->c[Z]==NULL)
      {
      printf (
         "ERROR (readrcv):  No room to allocate %i new (%i total) particles.\n",
         InputNP-a->np, InputNP);
      return EOF;
      }

   a->np = InputNP;


   /*
   Read particle types (these are individual bytes ->
   don't need  to reverse byte order due to endian mis-match)
   */
   FREAD(a->t[0], InputFile, a->np)

   /*  Find scale for particles  */
   LOOP (idir, NDIR)
      {
      if (Max[idir]==Min[idir])
         Scale[idir] = 0.0;
      else
         Scale[idir] = (Max[idir] - Min[idir]) / MAX_WORD;
      }


   /*  Read particle coordinates  */
   LOOP (ipart, a->np)
      {

      /*  Read scaled word coordinates  */
      //fread (WordCoord, sizeof(WORD2), 3, InputFile);
      FREAD (WordCoord, InputFile, 1)
      if (IsReverse)
         {
         REVERSE (WordCoord[X]);
         REVERSE (WordCoord[Y]);
         REVERSE (WordCoord[Z]);
         }

      /*  Scale to real coordinates  */
      LOOP (idir, NDIR)
         a->c[idir][ipart] = Scale[idir]*WordCoord[idir] + Min[idir];
      }

   /*  No errors - Return FALSE  */
   return 0;


   }
/*   ... ReadCorFile  */



/*
Open a coordinate file, use it analogously to fopen()

FileTypeStr can be either

"rt", "wt", "at"  - Read, Write or Append RCV (text)   file
"rb", "wb", "ab"  - Read, Write or Append COR (binary) file
"r",  "w" , "a"   - Read, Write or Append either file
                    (determine from file name extension)
*/
CFILE *OpenCoordinateFile (char *FileName, char *CFileTypeStr)
   {
   FILE  *FilePtr;
   CFILE *CFilePtr;
   int    CFileType;
   char   TempCFileTypeStr[3];

   /*  Check for illegal CFileTypeStr  */
   if (strlen(CFileTypeStr) < 1 || strlen(CFileTypeStr) > 2 )
      {
      printf 
			("INTERNAL ERROR:  (File %s Line %i) Illegal CFileTypeStr string (%s)",
			__FILE__,
			__LINE__,
         CFileTypeStr);
      printf (" sent to OpenCoordinateFile.\n");
      exit   (1);
      }

   /*  Copy Input file type string into buffer  */
   strncpy (TempCFileTypeStr, CFileTypeStr, 3);

   /*  Assign Coordinate file type from FileTypeStr  */
   if      (TempCFileTypeStr[1]=='t')
      CFileType=CF_RCV;
   else if (TempCFileTypeStr[1]=='b')
      CFileType=CF_COR;
   else
      CFileType=CF_NONE;

   /*  If no input file type specified, determine from file name.type  */
   if (CFileType==CF_NONE)
      {

      /*  Determine input file type  */
      CFileType = GetCoordFileType (FileName);
      if (CFileType==CF_NONE)
         {
         printf ("ERROR:  Input file type not recognized.  Must be\n");
         printf ("either COR or RCV.\n");
         exit   (1);
         }

      /*  Reset FileTypeStr  */
      switch (CFileType)
         {
         case CF_RCV:
            TempCFileTypeStr[1] = 't';
            break;
         case CF_COR:
            TempCFileTypeStr[1] = 'b';
         }

      /*  Make sure string terminator in place  */
      TempCFileTypeStr[2] = '\0';

      }

   /*  Open file  */
   FilePtr = fopen (FileName, TempCFileTypeStr);

   /*  Return null if file doesn't open  */
   if (FilePtr==NULL)
      return NULL;

   /*  Allocate new CFILE pointer  */
   CFilePtr = (CFILE *) calloc (1, sizeof(CFILE) );
   if (CFilePtr==NULL)
      return NULL;

   /*  Assign values to customized CFILE pointer  */
   CFilePtr->Ptr  = FilePtr;
   CFilePtr->Type = CFileType;

   return CFilePtr;
   }



/*  Returns TRUE if error  */
BOOLEAN ReadCoordinateFile (CFILE *CFilePtr, stepdata *CurStep)
   {
   BOOLEAN Status;

   /*  Read COR  format file  */
   if      (CFilePtr->Type==CF_COR)
      return ReadCorFile (CFilePtr->Ptr, CurStep);

   /*  Read RCV format file  */
   else if (CFilePtr->Type==CF_RCV)
      {

      Status = readrcv     (CFilePtr->Ptr, CurStep);

      /*  Convert units returned by readrcv() from ang to cm  */
      ScaleCoord (1e-8, CurStep);
      return Status;
      }

   /*  This statement is superfluous, but it avoids compilation warnings  */
   return TRUE;
   }


void WriteCoordinateFile
(
CFILE *CFilePtr,
stepdata *CurStep,
BOOLEAN SelFlag
)
   {

   if      (CFilePtr->Type==CF_COR)
      WriteCorFile (CFilePtr->Ptr, CurStep, SelFlag);

   else if (CFilePtr->Type==CF_RCV)
      writercv     (CFilePtr->Ptr, CurStep, SelFlag);

   }



void CloseCoordinateFile (CFILE *CFilePtr)
   {
   fclose (CFilePtr->Ptr);
   free   (CFilePtr);
   }



/*
************************************************************************
Local Functions
************************************************************************
*/

/*  Reverse bytes of word (works for any length WORD)  */
void ReverseWords (WORD *WordPtr, int NumWords)
   {
   BYTE *BytePtr;
   BYTE TempByte;
   int  FirstIndex;
   int  LastIndex;
   int  iword;

   /*  Set BYTE pointer to WORD array  */
   BytePtr = (BYTE *) WordPtr;

   LOOP (iword, NumWords)
      {

      /*  Find start and end of current word in list  */
      FirstIndex = iword * sizeof(WORD);
      LastIndex  = FirstIndex + sizeof(WORD) - 1;

      while (FirstIndex<LastIndex)
         {
         TempByte = BytePtr[FirstIndex];
         BytePtr[FirstIndex] = BytePtr[LastIndex];
         BytePtr[LastIndex ] = TempByte;
         FirstIndex++;
         LastIndex--;
         }

      }

   }
/* .. Reverse Word  */


/*  Read binary file into string until LineFeed encountered  */
void GetNextStringFromFile (char *InputString, int Length, FILE *InputFile)
   {
   int iread;
   char InputChar;

   iread = 0;
   do
      {
      //fread (&InputChar, 1, 1, InputFile);
      FREAD(InputChar, InputFile, 1)
      if (InputChar!='\r' && InputChar!='\n' && iread < Length-1)
         {
         InputString[iread] = InputChar;
         iread++;
         }
      }
      while (InputChar != '\r' && iread < Length-1);

   InputString[iread] = '\0';

   }

/*  Return time and date string  */
static char *GetCurrentTimeStr (void)
   {
   struct tm *timeptr;
   static char TimeString[23];
   time_t CurrentTime;

   /*  Get current time  (in seconds from 1900)  */
   time (&CurrentTime);

   /*  Convert to date, time  */
   timeptr = localtime (&CurrentTime);

   /*  Form string  */
   sprintf (TimeString, "%02d/%02d/%04d   %02d:%02d:%02d",
      timeptr->tm_mon+1, timeptr->tm_mday, 1900+timeptr->tm_year,
      timeptr->tm_hour, timeptr->tm_min, timeptr->tm_sec);

   /*  Return string  */
   return TimeString;
   }


/*  Return code for current file type (either CF_COR or CF_RCV)  */
int GetCoordFileType (char *FileName)
   {
   if (!strcmp(GetFileType(FileName),"COR"))
      return CF_COR;
   if (!strcmp(GetFileType(FileName),"RCV"))
      return CF_RCV;
   return CF_NONE;
   }


/*  Get file type from file name  */
char *GetFileType (char *FileName)
   {
   static char FileTypeBuffer[LEN_FILE_TYPE+1];
   int ichar;

   /*  Find period working from end of string  */
	ichar = strlen(FileName);
	while (ichar>=0 && FileName[ichar]!='.')
		ichar--;

	/*  No . suffix found  */
	if (ichar==-1)
		{
		FileTypeBuffer[0] = '\0';
		return FileTypeBuffer;
		}

   /*  Copy file type to buffer  */
   strncpy (FileTypeBuffer, FileName+ichar+1, LEN_FILE_TYPE+1);

   /*  Convert to uppercase  */
   LOOP (ichar, LEN_FILE_TYPE)
      {
      if (FileTypeBuffer[ichar]>='a' && FileTypeBuffer[ichar]<='z')
         FileTypeBuffer[ichar] += 'A' - 'a';
      }

   /*  Return pointer to buffer  */
   return FileTypeBuffer;
   }


/*  Reverse bytes of variable  */
void ReverseVar (void *VarPtr, int SizeVar)
   {
   BYTE *BytePtr;
   BYTE TempByte;
   int  FirstIndex;
   int  LastIndex;

   /*  Return if no bytes to switch  */
   if (SizeVar<2)
      return;

   /*  Set BYTE pointer to WORD array  */
   BytePtr = (BYTE *) VarPtr;

   /*  Find start and end of current word in list  */
   FirstIndex = 0;
   LastIndex  = FirstIndex + SizeVar - 1;

   while (FirstIndex<LastIndex)
      {
      TempByte = BytePtr[FirstIndex];
      BytePtr[FirstIndex] = BytePtr[LastIndex];
      BytePtr[LastIndex ] = TempByte;
      FirstIndex++;
      LastIndex--;
      }


   }
/* .. Reverse Var  */

void ReverseHeader (CorFileHeader_t *FileHeader)
   {
   REVERSE (FileHeader->Run)
   REVERSE (FileHeader->Step)
   REVERSE (FileHeader->Time)
   REVERSE (FileHeader->BoundaryType)
   REVERSE (FileHeader->Surf[X])
   REVERSE (FileHeader->Surf[Y])
   REVERSE (FileHeader->Surf[Z])
   REVERSE (FileHeader->Box[X])
   REVERSE (FileHeader->Box[Y])
   REVERSE (FileHeader->Box[Z])
   REVERSE (FileHeader->Etot)
   REVERSE (FileHeader->Epot)
   REVERSE (FileHeader->Ekin)
   REVERSE (FileHeader->Ebath)
   REVERSE (FileHeader->Min[X])
   REVERSE (FileHeader->Min[Y])
   REVERSE (FileHeader->Min[Z])
   REVERSE (FileHeader->Max[X])
   REVERSE (FileHeader->Max[Y])
   REVERSE (FileHeader->Max[Z])
   REVERSE (FileHeader->Np)
   }


void CalcMinMaxCoord
(
stepdata *a,
BOOLEAN UseSelect,
BOOLEAN UseBoundary
)
   {
   int FirstPart;
   int idir;
   int ipart;
   float TempCoord;


   /*  Find first selected particle  */
   FirstPart = 0;
   if (UseSelect)
      while (!a->sel[FirstPart] && FirstPart<a->np)
         FirstPart++;

   /*  Test for no selected partices, then use all  */
   if (FirstPart==a->np)
      {
      FirstPart = 0;
      UseSelect = FALSE;
      }

   LOOP (idir, NDIR)
      {

      /*  Check boundary conditions  */
      if (UseBoundary && a->boundrep[idir])
         {
         a->MinCoord[idir] = 0.0;
         a->MaxCoord[idir] = a->box[idir];
         }

      /*  use particles positions  */
      else
         {
         /*  Initialize min and maximum particle  */
         a->MinCoord[idir] = a->MaxCoord[idir] = a->c[idir][FirstPart];

         /*  Search for max amd min coordinate in each direction  */
         for (ipart=FirstPart+1; ipart<a->np; ipart++)
            if (!UseSelect || a->sel[ipart])
               {
               TempCoord = a->c[idir][ipart];
               if (TempCoord < a->MinCoord[idir])
                  a->MinCoord[idir] = TempCoord;
               else if (TempCoord > a->MaxCoord[idir])
                  a->MaxCoord[idir] = TempCoord;
               }
         }
      }
   }


void ScaleCoord (double Scale, stepdata *Step)
   {
   int idir;
   int ipart;

   LOOP (idir, NDIR)
   LOOP (ipart, Step->np)
      if (Step->c[idir] != NULL)
         {
         Step->c[idir][ipart] *= Scale;
         }

   LOOP (idir, NDIR)
      {
      Step->box[idir] *= Scale;
      }

   }
