/*
************************************************************************
History
************************************************************************
*/

/*
3 May 1996
	Correct calculation of DisplayedDisp[X][] etc, was not taking wrapping
	into account.
10 May 1996
	Fixed calculation of Disp[HOR], Disp[VER].  Was using wrong index
	for CoordCur[HOR][]...

11 July 1996
	- Add ability to view particles from arbitrary direction
	- Combine scalar triplets or pairs into arrays indexed by
	  X,Y,Z or HOR,VER,DEP

11 Mar 1997
   - For PlotParam->FixedFrame option, check particles
    to make sure they are within frame before plotting.
*/

/*
************************************************************************
Include Files
************************************************************************
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <bondsub.h>
#include <sortsub.h>
#include <math.h>
#include "drawstep.h"
#include "hlgraph.h"
#include "llgraph.h"


/*
************************************************************************
Compile Switches
************************************************************************
*/
#define SVGA
#undef  SVGA

#define EGA

#define DEBUG
#undef  DEBUG



/*
************************************************************************
Defines
************************************************************************
*/

/*  Use far only for PC  */
#ifndef NOGRAPHICS
#ifdef __TURBOC__
#define FAR far
#else
#define FAR
#endif
#endif


/*  Used for screen directions  */
#define HOR 0
#define VER 1
#define DEP 2

/*  Number of corners on containing box  */
#define NUM_CORNER 8

/*  Default min radius for slip neighbor search  */
#define	DEFAULT_MIN_RADIUS	0.0

#ifndef M_PI
#define M_PI		  3.14159265358979323846	/* pi */
#endif

/*  Default Atom Radius Squared  */
#define DEFAULT_ATOM_RADIUS_SQR 2.5


/*
************************************************************************
Macros
************************************************************************
*/

/*  Print out current file and line number  */
#define WRITEMSG \
	printf ("In file %s at line %i.\n", __FILE__, __LINE__);

/*  Write out variable name and value	*/
#define WRITEVAR(VAR,TYPE) \
	printf ("File %s  Line %i : ", __FILE__, __LINE__); \
	printf ("%s = ", #VAR); \
	printf (#TYPE, VAR); \
	printf ("\n");

#define ALLOCATE(VAR,TYPE,NUM) \
	VAR = (TYPE *) calloc ( (NUM), sizeof ( TYPE ) ); \
	if (VAR == NULL) \
		{ \
		ShutdownGraphics(); \
		printf ("ERROR (File %s line %i)\n", __FILE__, __LINE__); \
		printf ("   Cannot allocate variable %s.\n", #VAR); \
		}

#define FREE(VAR) \
	if ((VAR)!=NULL) \
		{ \
		free (VAR); \
		(VAR) = NULL; \
		}

#define LOOP(INDEX,LIMIT) \
	for (INDEX=0; INDEX<(LIMIT); INDEX++)

#define WRAP(COORD,DIR) \
	if (CurStep->boundrep[DIR])  \
		{	\
		if (COORD < -HalfBox[DIR])  \
			COORD += CurStep->box[DIR];  \
		else if (COORD > HalfBox[DIR])  \
			COORD -= CurStep->box[DIR];  \
		}


/*  Rotate particles  (stored in 3 separate arrays [x,x,..] [y,y,..] etc.)  */
#define ROTATE_PART(B,M,A,I) \
	{	\
	B[X]= M[X][X]*A[X][I]+ M[X][Y]*A[Y][I]+ M[X][Z]*A[Z][I]; \
	B[Y]= M[Y][X]*A[X][I]+ M[Y][Y]*A[Y][I]+ M[Y][Z]*A[Z][I]; \
	B[Z]= M[Z][X]*A[X][I]+ M[Z][Y]*A[Y][I]+ M[Z][Z]*A[Z][I]; \
	}

/*  Rotate coordinates (stored in 1 combined array [x,y,z,x,y,z...])  */
#define ROTATE_COORD(B,M,A) \
	{	\
	B[X]= M[X][X]*A[X]+ M[X][Y]*A[Y]+ M[X][Z]*A[Z]; \
	B[Y]= M[Y][X]*A[X]+ M[Y][Y]*A[Y]+ M[Y][Z]*A[Z]; \
	B[Z]= M[Z][X]*A[X]+ M[Z][Y]*A[Y]+ M[Z][Z]*A[Z]; \
	}


/*  Assign first number to Min (or Max)  */
#define MIN(A,B)	 if (B<A) A=B;
#define MAX(A,B)	 if (B>A) A=B;


/*
************************************************************************
Module Wide Variables
************************************************************************
*/

/*  Display colors and default values  */
WORD2 TextColor_m  =  1;
WORD2 FrameColor_m =  1;
WORD2 BondColor_m  =  1;
WORD2 DispColor_m  =  1;
WORD2 AtomColor_m  =  1;

/*  Variables used by BondCallBack()  */
BOOLEAN ColorBondByDistFlag_m;
float   BondFactor_m;
float   MinBond_m;
WORD2 	*BondColorMap_m;

/*  Variables used by SlipCallBack()  */
static int	*ImageCur_m 	[NDIR];
static int	*ImageDisp_m	[NDIR];
static int	*SumImageDisp_m[NDIR];
static int	*CntImageDisp_m;


/*
************************************************************************
Local Function Prototypes
************************************************************************
*/
void SlipCallBack (int FirstIndex, int SecondIndex, float Radius);
void NormalizeSlipDisplacements
	( stepdata *, unsigned *, int, float, int *[NDIR]);
void BondCallBack (int FirstIndex, int SecondIndex, float Radius);

void ShutdownGraphics (void);

void CalcRotatedLimits (float[2][NDIR], stepdata *, dsparm *);

BOOLEAN AreIdenticalLimits (float MinLimits[NDIR], float MaxLimits[NDIR]);

void ClipPoint (int, int, int *, int *, int, int, int, int);


/*
************************************************************************
Exported functions
************************************************************************
*/

int DS_DrawStep
(
stepdata  *RefStep,
stepdata  *CurStep,
dsparm	 *PlotParam
)
	{
	float ImageRange[NDIR];
	float RotatedCoordRange[NDIR];
	float CoordAspectRatio;
	float ImageAspectRatio;
	float ScaleCoordToImage[NDIR];
	float ScaleCoordToLayer;
	float RotatedCoordOffset [NDIR];
	float ScreenOffset[NDIR];
	int	FrameMin[NDIR];
	int	FrameMax[NDIR];
	int	FrameRange[NDIR];
	float *CoordCur[NDIR];
	float *CoordRef[NDIR];
	float Disp[NDIR];
	float RotatedCoord[NDIR];
	float RotatedDisp[NDIR];
	float Angle;
	int	*ImageCur[NDIR];
	int	*ImageDisp[NDIR];
	int	i;
	int	isort;
	int	icoord;
	int	image;
	int	NumImagePart;
	float *DisplayedDisp[NDIR];
	float DotProd;
	float ColorFactor;
	float DegreesFromRadians;
	WORD2  TempColor;
	double TempRadius;
	WORD2  DispIndex;
	int	 LayerIndex;
	WORD2  TypeIndex;
	float BondRange;
	float RotatedCoordLimits[2][NDIR];
	float LayerThickness;
	int   TailEndHor;
	int   TailEndVer;

	/*  Indices of atoms in selected layer sorted by depth  */
	unsigned *IxSortToImage;

	/*  Holds index of original particle for atoms in layer	*/
	unsigned *IxImageToCoord;
	char IndexString[32];
	int  StringLength;
	int  xpos;
	int  ypos;
	float HalfBox[NDIR];
	float MaxMagnitude = -999.9, MinMagnitude = 999.9, CurMag;
	float MinBound, MaxBound;
	float MinDotProd= 999.9, MaxDotProd=-999.9;


	/*   Box size	*/
	HalfBox[X] = CurStep->box[X]/2;
	HalfBox[Y] = CurStep->box[Y]/2;
	HalfBox[Z] = CurStep->box[Z]/2;

	/*   Calculate rotated limits  */
	CalcRotatedLimits (RotatedCoordLimits, CurStep, PlotParam);
	RotatedCoordRange [HOR] = RotatedCoordLimits[1][HOR] - RotatedCoordLimits[0][HOR];
	RotatedCoordRange [VER] = RotatedCoordLimits[1][VER] - RotatedCoordLimits[0][VER];
	RotatedCoordRange [DEP] = RotatedCoordLimits[1][DEP] - RotatedCoordLimits[0][DEP];
	RotatedCoordOffset[HOR] = RotatedCoordLimits[0][HOR];
	RotatedCoordOffset[VER] = RotatedCoordLimits[0][VER];
	RotatedCoordOffset[DEP] = RotatedCoordLimits[0][DEP];


	/*  Size of on-screen frame  */
	FrameRange  [HOR] = PlotParam->hsmax - PlotParam->hsmin;
	FrameRange  [VER] = PlotParam->vsmax - PlotParam->vsmin;

	/*  On-screen offset  */
	ScreenOffset[HOR] = PlotParam->hsmin + PlotParam->limarg;
	ScreenOffset[VER] = PlotParam->vsmin + PlotParam->timarg;
	ScreenOffset[DEP] = 0.0;

	/*  Reserve room for label  */
	if (PlotParam->LabelFlag)
		{
		FrameRange  [HOR] -= PlotParam->lmarg + PlotParam->rmarg;
		FrameRange  [VER] -= PlotParam->tmarg + PlotParam->bmarg;
		ScreenOffset[HOR] += PlotParam->lmarg;
		ScreenOffset[VER] += PlotParam->tmarg;
		}

	/*  Size of on-screen Image  (smaller than frame)	*/
	ImageRange[HOR] = FrameRange[HOR] - PlotParam->limarg - PlotParam->rimarg;
	ImageRange[VER] = FrameRange[VER] - PlotParam->bimarg - PlotParam->timarg;



	/*  DETERMINE IF PICTURE BOUNDED IN HORIZONTAL OR VERTICAL DIRECTION  */
	CoordAspectRatio = RotatedCoordRange[VER] / RotatedCoordRange[HOR];
	ImageAspectRatio = PlotParam->aspect * ImageRange[VER] / ImageRange[HOR];

	/*  BOUNDED IN VERTICAL   DIRECTION  */
	if ( CoordAspectRatio > ImageAspectRatio )
		{

		/*  Determine mapping (scale) from Coordinates to Image	*/
		ScaleCoordToImage[VER] = ImageRange[VER] / RotatedCoordRange[VER];
		ScaleCoordToImage[HOR] = ScaleCoordToImage[VER]  * PlotParam->aspect;
		ScaleCoordToImage[DEP] = ScaleCoordToImage[VER];

		/*  Adjust image offsets to center image	*/
		ScreenOffset[HOR] +=
			(ImageRange[HOR] - ScaleCoordToImage[HOR]*RotatedCoordRange[HOR]) / 2;
		FrameRange[HOR] =
			ScaleCoordToImage[HOR] * RotatedCoordRange[HOR] +
			PlotParam->limarg + PlotParam->rimarg;
		}

	/*  BOUNDED IN HORIZONTAL DIRECTION  */
	else
		{
		/*  DETERMINE SCALE	*/
		ScaleCoordToImage[HOR] = ImageRange[HOR] / RotatedCoordRange[HOR];
		ScaleCoordToImage[VER] = ScaleCoordToImage[HOR]  / PlotParam->aspect;
		ScaleCoordToImage[DEP] = ScaleCoordToImage[VER];
		/* ADJUST OFFSET TO CENTER IMAGE */
		ScreenOffset[VER] +=
			(ImageRange[VER] - ScaleCoordToImage[VER]*RotatedCoordRange[VER]) / 2;
		FrameRange[VER] =
			ScaleCoordToImage[VER] * RotatedCoordRange[VER] +
			PlotParam->timarg + PlotParam->bimarg;
		}

	/*  Correct values to place origin at bottom-left instead of top-left  */
	ScreenOffset[VER] 	  +=	ScaleCoordToImage[VER]*RotatedCoordRange[VER];
	ScaleCoordToImage[VER]	= -ScaleCoordToImage[VER];

	/*  Calculate image coordinate of frame corners  */
	FrameMin[HOR] = ScreenOffset[HOR]  - PlotParam->limarg;
	FrameMax[VER] = ScreenOffset[VER]  + PlotParam->bimarg;
	FrameMax[HOR] = FrameMin[HOR] + FrameRange[HOR];
	FrameMin[VER] = FrameMax[VER] - FrameRange[VER];

	/*  Draw frame  */
	if (FrameColor_m!=CLEAR_COLOR)
		{
		GR_Hor (FrameMin[VER], FrameMin[HOR], FrameMax[HOR], FrameColor_m);
		GR_Hor (FrameMax[VER], FrameMin[HOR], FrameMax[HOR], FrameColor_m);
		GR_Ver (FrameMin[HOR], FrameMin[VER], FrameMax[VER], FrameColor_m);
		GR_Ver (FrameMax[HOR], FrameMin[VER], FrameMax[VER], FrameColor_m);
		}


	/*
	***********************************************
	convert float coordinates to screen coordinates
	***********************************************
	*/

	/*  If no selected coordinates then return no error  */
	if (CurStep->nsel==0)
		return 0;

	/*  Allocate storage  */
	if (RefStep)
		{
		ALLOCATE ( ImageDisp[HOR],   int,  CurStep->nsel)
		ALLOCATE ( ImageDisp[VER],   int,  CurStep->nsel)
		}
	ALLOCATE ( ImageCur[HOR],	int,	  CurStep->nsel)
	ALLOCATE ( ImageCur[VER],	int,	  CurStep->nsel)
	ALLOCATE ( ImageCur[DEP],	int,	  CurStep->nsel)
	ALLOCATE ( DisplayedDisp[X],	float,  CurStep->nsel)
	ALLOCATE ( DisplayedDisp[Y],	float,  CurStep->nsel)
	ALLOCATE ( DisplayedDisp[Z],	float,  CurStep->nsel)

	/*  Get pointers to horizontal, vertical and depth coordinates  */
	CoordCur[X] = CurStep->c[X];
	CoordCur[Y] = CurStep->c[Y];
	CoordCur[Z] = CurStep->c[Z];
	if (RefStep)
		{
		CoordRef[X] = RefStep->c[X];
		CoordRef[Y] = RefStep->c[Y];
		CoordRef[Z] = RefStep->c[Z];
		}

	/*  Allocate space for atom indices  */
	ALLOCATE ( IxImageToCoord, unsigned,	CurStep->nsel)
	ALLOCATE ( IxSortToImage,	unsigned,	CurStep->nsel)



	/*
	- Choose particle to display
   - Store their image coordinates
	*/
	NumImagePart = 0;
	LOOP (i, CurStep->np)
		{

		/*  Reject current particle if not selected or already too many of them  */
		if ( !CurStep->sel[i] || NumImagePart>=CurStep->nsel)
			continue;

		/*  
		Reject current particle if using fixed frame 
		and particles are outside frame
		*/
		if 
		(
		PlotParam->FixedFrameFlag &&
			(
			CoordCur[HOR][i]< PlotParam->Frame[0][HOR] || 
			CoordCur[HOR][i]> PlotParam->Frame[1][HOR] ||
			CoordCur[VER][i]< PlotParam->Frame[0][VER] || 
			CoordCur[VER][i]> PlotParam->Frame[1][VER] ||
			CoordCur[DEP][i]< PlotParam->Frame[0][DEP] || 
			CoordCur[DEP][i]> PlotParam->Frame[1][DEP] 
			)
		)
			continue;

		/*  Get Rotated Coordinates  */
		ROTATE_PART(RotatedCoord, PlotParam->Rotate, CoordCur, i)

		ImageCur[HOR][NumImagePart] =
			ScaleCoordToImage[HOR] * (RotatedCoord[HOR]-RotatedCoordOffset[HOR]) +
			ScreenOffset[HOR];
		ImageCur[VER][NumImagePart] =
			ScaleCoordToImage[VER] * (RotatedCoord[VER]-RotatedCoordOffset[VER]) +
			ScreenOffset[VER];
		ImageCur[DEP][NumImagePart] =
			ScaleCoordToImage[DEP] * (RotatedCoord[DEP]-RotatedCoordOffset[DEP]) +
			ScreenOffset[DEP];

		/*  Determine displacement info	*/
		if (RefStep)
			{
			/*  Calculate particle displacement  */
			Disp[X] = CoordRef[X][i] - CoordCur[X][i];
			Disp[Y] = CoordRef[Y][i] - CoordCur[Y][i];
			Disp[Z] = CoordRef[Z][i] - CoordCur[Z][i];

			/*  Take repeating boundaries into account  */
			WRAP (Disp[HOR], HOR)
			WRAP (Disp[VER], VER)
			WRAP (Disp[DEP], DEP)

			/*
			Only for displayed particles
			save displacement in original orientation
			*/
			DisplayedDisp[X][NumImagePart] = Disp[X];
			DisplayedDisp[Y][NumImagePart] = Disp[Y];
			DisplayedDisp[Z][NumImagePart] = Disp[Z];

			/*  Get Rotated Displacement	*/
			ROTATE_COORD(RotatedDisp, PlotParam->Rotate, Disp)


			/*  Calculate displacement in pixels  */
			ImageDisp[HOR][NumImagePart] = ScaleCoordToImage[HOR] * RotatedDisp[HOR];
			ImageDisp[VER][NumImagePart] = ScaleCoordToImage[VER] * RotatedDisp[VER];

			/*
			Figure minimum and maximum displacment (and their dot products)
			for use later in scaling
			*/
			CurMag =
				DisplayedDisp[X][NumImagePart] * DisplayedDisp[X][NumImagePart] +
				DisplayedDisp[Y][NumImagePart] * DisplayedDisp[Y][NumImagePart] +
				DisplayedDisp[Z][NumImagePart] * DisplayedDisp[Z][NumImagePart];
			CurMag = sqrt(CurMag);

			if (CurMag<MinMagnitude)
				MinMagnitude = CurMag;

			if (CurMag>MaxMagnitude)
				MaxMagnitude = CurMag;

			DotProd = (DisplayedDisp[X][NumImagePart] * PlotParam->vector[X]) +
						 (DisplayedDisp[Y][NumImagePart] * PlotParam->vector[Y]) +
						 (DisplayedDisp[Z][NumImagePart] * PlotParam->vector[Z]);

			if (DotProd < MinDotProd)
				MinDotProd = DotProd;
			if (DotProd > MaxDotProd)
				MaxDotProd = DotProd;
			}

		/*  Store index of original atom for each atom in current layer  */
		IxImageToCoord[NumImagePart] = i;
		NumImagePart++;
		
		}
		/*  Calculate image coordinates  */

	if ( (RefStep) && ((MaxMagnitude < -999) || (MinMagnitude > 999)))
		{
		printf("Uh oh..  Min or  Max Magnitude is not set low or high enough.\n");
		printf("The colors would be all messed up.  About line 463 in drawstep.c\n");
		printf(" %f %f ",MinMagnitude, MaxMagnitude);
		exit (-1);
		}

	/*  Color atoms by displacmenet magnitude according to fixed scale  */
	if
	(
	PlotParam->ColorAtomByDispMagFlag ||
	PlotParam->ColorAtomByInplaneDispMagFlag
	)
		{
		MinBound = 0.0;
		MaxBound = PlotParam->DispMag;
		if (MaxBound==0) { printf("MAXBOUND\n");exit(1);}
		ColorFactor   =
			(float) (PlotParam->NumDispColor - 1) / (MaxBound - MinBound);
		}

	/*  Color atoms by displacmenet magnitude according to rel. scale  */
	else if
	(
	PlotParam->ColorAtomByRelDispMagFlag ||
	PlotParam->ColorAtomByInplaneRelDispMagFlag
	)
		{
		MinBound = MinMagnitude;
		MaxBound = MaxMagnitude;
		ColorFactor   =
			(float) (PlotParam->NumDispColor) / (MaxBound-MinBound);
		}

	/*  Color atoms by displacmenet dot prod. according to fixed scale  */
	else if (PlotParam->ColorAtomByDispVectFlag)
		{
		MinBound = PlotParam->vectmin;
		MaxBound = PlotParam->vectmax;
		ColorFactor   =
			(float) (PlotParam->NumDispColor - 2) / (MaxBound-MinBound);
		}

	/*  Color atoms by displacmenet dot prod. according to rel.  scale  */
	else if (PlotParam->ColorAtomByRelDispVectFlag)
		{
		MinBound = MinDotProd;
		MaxBound = MaxDotProd;
		ColorFactor   =
			(float) (PlotParam->NumDispColor) / (MaxBound - MinBound);
		}

	/*  Color atoms by angle within display plane  */
	else if (PlotParam->ColorAtomByDispAngleFlag)
		{
		MinBound = PlotParam->DispMag;
		ColorFactor   =
			(float) (PlotParam->NumDispColor - 1) / 360;
		DegreesFromRadians = 180.0 / M_PI;
		}

	else if (PlotParam->ColorAtomByLayerFlag)
		{
		CalcMinMaxCoord (CurStep, TRUE, TRUE);
		LayerThickness = 
			CurStep->MaxCoord[DEP] - CurStep->MinCoord[DEP];
		if (LayerThickness==0.0)
			{
			ScaleCoordToLayer = 0;
			}
		else
			{
			ScaleCoordToLayer =
				PlotParam->NumLayerColor /
				(CurStep->MaxCoord[DEP] - CurStep->MinCoord[DEP]);
			}
		}


	/*  If doing slip tails then normalize displacements  */
	if (PlotParam->TailSlipFlag && RefStep!=NULL)
		NormalizeSlipDisplacements
			(
			CurStep,
			IxImageToCoord,
			NumImagePart,
			PlotParam->TailSlipRad,
			ImageDisp
			);

	/*  
	If not using a fixed frame, then number of image particles should
   equal number of selected particles
	*/
	if ( NumImagePart != CurStep->nsel && !PlotParam->FixedFrameFlag)
		{
		fprintf (stderr,
			"ERROR (drawstep):  ");
		fprintf (stderr,
			"Internal inconsistency between counts of selected atoms.\n");
		return 1;
		}

	/*  SORT SCREEN COORDINATES BY DEPTH  */
	LOOP (i, NumImagePart)
		IxSortToImage[i] = i;
	sortix (ImageCur[DEP], IxSortToImage, NumImagePart);


	/*  Initialize bond calculation	*/
	ColorBondByDistFlag_m = PlotParam->ColorBondByDistFlag;
	if (ColorBondByDistFlag_m)
		{
		MinBond_m				 = PlotParam->bnd1;
		BondColorMap_m 		 = PlotParam->BondColorMap;
		BondRange				 = (PlotParam->bnd2 - PlotParam->bnd1);
		if (BondRange==0)
			BondFactor_m = 0;
		else
			BondFactor_m = PlotParam->NumBondColor / BondRange;
		}


	/*
	****************************
	Draw Atoms and Displacements
	****************************
	*/

	if (PlotParam->AtomFlag || PlotParam->TailFlag)
	LOOP (isort, NumImagePart)
		{
		image  = IxSortToImage [isort];
		icoord = IxImageToCoord[image];

		/*
		*********************
		Determine Atom Radius
		*********************
		*/
		if (PlotParam->RadiusAtomByTypeFlag)
			{
			TypeIndex  = CurStep->t[icoord] % MAX_COLOR_MAP;
			TempRadius = PlotParam->AtomRadiusMap[TypeIndex];
			}

		/*  Use default radius  */
		else
			{
			TempRadius = DEFAULT_ATOM_RADIUS_SQR;
			}


		/*
		********************
		Determine Atom Color
		********************
		*/

		/*  Color atoms by type  */
		if (PlotParam->ColorAtomByTypeFlag)
			{
			TypeIndex = CurStep->t[icoord] % MAX_COLOR_MAP;
			TempColor = PlotParam->AtomColorMap[TypeIndex];
			}


		/*  Color atoms by displacement magnitude */
		else if
		(
		PlotParam->ColorAtomByDispMagFlag ||
		PlotParam->ColorAtomByRelDispMagFlag
		)
			{

			/*
			If there is no reference step for calculating displacment,
			assume displacement is zero
			*/

			if (RefStep==NULL)
				CurMag = 0;

			else
				CurMag = sqrt( DisplayedDisp[X][image] * DisplayedDisp[X][image]
								+ DisplayedDisp[Y][image] * DisplayedDisp[Y][image]
								+ DisplayedDisp[Z][image] * DisplayedDisp[Z][image]);

			DispIndex = (WORD2) ( ColorFactor * (CurMag - MinBound) );

			if (DispIndex >= PlotParam->NumDispColor)
				DispIndex = PlotParam->NumDispColor-1;

			TempColor = PlotParam->AtomColorMap[DispIndex];
			}

		/*  Color atoms by displacement dot product	*/
		else if
		(
		PlotParam->ColorAtomByDispVectFlag ||
		PlotParam->ColorAtomByRelDispVectFlag
		)
			{

			/*
			If there is no reference step for calculating displacment,
			assume displacement is zero
			*/
			if (RefStep==NULL)
				DotProd = 0.0;

			else
				DotProd = (DisplayedDisp[X][image] * PlotParam->vector[X]) +
							 (DisplayedDisp[Y][image] * PlotParam->vector[Y]) +
							 (DisplayedDisp[Z][image] * PlotParam->vector[Z]);

			DispIndex = (WORD2) ( ColorFactor * (DotProd - MinBound) );

			/*
			If doing absolute dot product scaling,
			correct DispIndex for below or above range
			*/
			if (PlotParam->ColorAtomByDispVectFlag)
				{

				/*  Dot product below range, set at first color in list	*/
				if (DotProd<MinBound)
					DispIndex = 0;

				/*  Dot product above range, set at last	color in list	*/
				else if (DotProd>MaxBound)
					DispIndex = PlotParam->NumDispColor-1;

				/*  Dot product in range, map to colors between first and last  */
				else
					DispIndex++;
				}

			/*
			If doing relative dot product scaling
			correct for out of range color due to overflow
			*/
			if (DispIndex>=PlotParam->NumDispColor)
				DispIndex = PlotParam->NumDispColor-1;


			TempColor = PlotParam->AtomColorMap[DispIndex];
			}


		/*  Color atoms by inplane displacement magnitude	*/
		else if
		(
		PlotParam->ColorAtomByInplaneDispMagFlag ||
		PlotParam->ColorAtomByInplaneRelDispMagFlag
		)
			{
			/*
			If there is no reference step for calculating displacment,
			use min displacment value
			*/
			if (RefStep==NULL)
				CurMag = 0.0;

			else
				{

				/*  Calculate displacement magnitude in display plane */
				Disp[X] = DisplayedDisp[X][image];
				Disp[Y] = DisplayedDisp[Y][image];
				Disp[Z] = DisplayedDisp[Z][image];

				ROTATE_PART(RotatedDisp, PlotParam->Rotate, DisplayedDisp, image)

				CurMag =
					sqrt
					(
					RotatedDisp[HOR] * RotatedDisp[HOR] +
					RotatedDisp[VER] * RotatedDisp[VER]
					);

				}

			DispIndex = (WORD2) ( ColorFactor * (CurMag - MinBound) );

			if (DispIndex >= PlotParam->NumDispColor)
				DispIndex = PlotParam->NumDispColor-1;

			TempColor = PlotParam->AtomColorMap[DispIndex];

			}

		/*  Color atom by displacement angle  */
		else if (PlotParam->ColorAtomByDispAngleFlag)
			{

			/*
			If there is no reference step for calculating displacment,
			use min displacment value
			*/
			if (RefStep==NULL)
				TempColor = PlotParam->AtomColorMap[0];

			else
				{

				/*  Calculate displacement magnitude in display plane */
				ROTATE_PART(RotatedDisp, PlotParam->Rotate, DisplayedDisp, image)

				CurMag =
					sqrt
					(
					RotatedDisp[HOR] * RotatedDisp[HOR] +
					RotatedDisp[VER] * RotatedDisp[VER]
					);

				/*  If magnitude too small, color with first color  */
				if (CurMag<MinBound)
					TempColor = PlotParam->AtomColorMap[0];

				else
					{

					/*
					Calculate angle of displacement
					  - invert and add pi so that angle range from 0 to 2*pi
					  - add input phase
					*/
					Angle =
						DegreesFromRadians *
						atan2 (-RotatedDisp[HOR], RotatedDisp[VER]) +
						180 +  PlotParam->Phase;

					/*  Wrap angle into range	*/
					Angle = fmod (Angle, 360.0);

					/*  Set color acording to angle	*/
					DispIndex = ColorFactor * Angle + 1;
					TempColor = PlotParam->AtomColorMap[DispIndex];


					}

				}

			}


		/*  else color atom by layer	*/
		else if (PlotParam->ColorAtomByLayerFlag)
			{
			LayerIndex =  ScaleCoordToLayer *
				CoordCur[DEP][icoord] - CurStep->MinCoord[DEP];
			TempColor = PlotParam->AtomColorMap[LayerIndex];
			}


		/*  Use text color  */
		else
			TempColor = TextColor_m;



		/*
		*********
		Draw Atom
		*********
		*/

		/*  If color is clear, dont' draw  */
		if (TempColor!=CLEAR_COLOR)
			{

			/*  Do not draw atom  */
			if (!PlotParam->AtomFlag)
				{
				}

			/*  Write out index number for atom ...	*/
			else if (PlotParam->AtomIndexFlag)
				{
				sprintf (IndexString, "%i", icoord+1);
				StringLength = strlen(IndexString);
				GR_Printf
					(
					ImageCur[HOR][image] - StringLength*4,
					ImageCur[VER][image]-4,
					TempColor,
					IndexString
					);
				}

			/*  ... else use particle symbol  */
			else
				{
				icoord = IxImageToCoord[image];
				xpos = ImageCur[HOR][image];
				ypos = ImageCur[VER][image];

				GR_Atom (xpos, ypos, TempColor, TempRadius);
				}


			/*
			******************
			Draw displacements
			******************
			*/

			if (PlotParam->TailFlag && RefStep!=NULL)
				{

				/*  Skip if no tail  */
				if (ImageDisp[HOR][image]==0 && ImageDisp[VER][image]==0)
					continue;

				/*
				If color atoms by layer, color displacement by ref position
				*/
				if (PlotParam->ColorAtomByLayerFlag)
					{
					/*
					Insure that Reference coordinate does not exceed
					current coordinate limits
					*/
					LayerIndex =  ScaleCoordToLayer *
						CoordRef[DEP][icoord] - CurStep->MinCoord[DEP];
					if (LayerIndex<0)
						LayerIndex = 0;
					else if (LayerIndex>=PlotParam->NumLayerColor)
						LayerIndex = PlotParam->NumLayerColor;
					TempColor = PlotParam->AtomColorMap[LayerIndex];
					}

				TailEndHor = 
					ImageCur[HOR][image] + 
					PlotParam->TailScale * ImageDisp[HOR][image];
				TailEndVer = 
					ImageCur[VER][image] + 
					PlotParam->TailScale * ImageDisp[VER][image];
				ClipPoint 
					(
					ImageCur[HOR][image],
					ImageCur[VER][image],
					&TailEndHor,
					&TailEndVer,
					FrameMin[HOR]+1,
					FrameMin[VER]+1,
					FrameMax[HOR]-1,
					FrameMax[VER]-1
					);
				GR_Line 
					(
					ImageCur[HOR][image], 
					ImageCur[VER][image], 
					TailEndHor, 
					TailEndVer, 
					TempColor
					);
#if 0
				GR_Line
					(
					ImageCur[HOR][image],
					ImageCur[VER][image],
					ImageCur[HOR][image] + PlotParam->TailScale * ImageDisp[HOR][image],
					ImageCur[VER][image] + PlotParam->TailScale * ImageDisp[VER][image],
					TempColor
					);
#endif
				}


			}	/*  TempColor!=CLEAR_COLOR  */


		}	/*  Draw atoms and tails */






	/*
	**********
	Draw Bonds
	**********
	*/

	ImageCur_m[HOR]  = ImageCur[HOR];
	ImageCur_m[VER]  = ImageCur[VER];
	if (NumImagePart>0)
		{
		WORD2 boundrep[NDIR];

		if (PlotParam->BondWrapFlag)
			memcpy (boundrep, CurStep->boundrep, sizeof(boundrep) );
		else
			memset (boundrep, 0, 					 sizeof(boundrep) );

		bondx
			(
			CurStep->c,
			IxImageToCoord,
			NumImagePart,
			PlotParam->bnd1,
			PlotParam->bnd2,
			CurStep->box,
			(unsigned *) boundrep,
			BondCallBack
			);

		}

	/*  RELEASE STORAGE	*/
	FREE (IxSortToImage);
	FREE (IxImageToCoord);
	FREE (ImageCur[DEP]);
	FREE (ImageCur[VER]);
	FREE (ImageCur[HOR]);
	FREE (DisplayedDisp[X]);
	FREE (DisplayedDisp[Y]);
	FREE (DisplayedDisp[Z]);
	if (RefStep)
		{
		FREE (ImageDisp[VER]);
		FREE (ImageDisp[HOR]);
		}


	/*  Return No Error	*/
	return 0;
	}


void DS_DrawLabel (stepdata *p, dsparm *dp)
	{
	WORD2 LowerX;
	WORD2 UpperX;

	UpperX = 0;
	LowerX = dp->vsmax - 8;

	/*  Do not draw labels if text color is clear  */
	if (TextColor_m==CLEAR_COLOR)
		return;

	GR_Printf (  0,UpperX, TextColor_m,
		"Run %li",  p->run);
	GR_Printf (100,UpperX, TextColor_m,
		"Step %li", p->step);
	GR_Printf (300,UpperX, TextColor_m,
		"Number %i of %i", p->nsel, p->np);
	GR_Printf (  0,LowerX, TextColor_m,
		"Temp %10.4f", p->tempc);
	GR_Printf (300,LowerX, TextColor_m,
		"Layer %i of %i", dp->ilayer, dp->nlayer);
	}

void DS_SetFrameColor (WORD2 InputColor)
	{
	FrameColor_m = InputColor;
	}

void DS_SetTextColor (WORD2 InputColor)
	{
	TextColor_m = InputColor;
	}

void DS_SetDispColor (WORD2 InputColor)
	{
	DispColor_m = InputColor;
	}

void DS_SetBondColor (WORD2 InputColor)
	{
	BondColor_m = InputColor;
	}


/*
************************************************************************
Local Functions
************************************************************************
*/

/*  BOND SUBROUTINE FUNCTION	*/
#pragma argsused
void BondCallBack (int FirstIndex, int SecondIndex, float RadiusSqr)
	{
	float Radius;
	int	Index;
	WORD2  TempColor;

	/*  If coloring bonds  */
	if (ColorBondByDistFlag_m)
		{
		Radius	 = sqrt(RadiusSqr);
		Index 	 = BondFactor_m * (Radius-MinBond_m);
		TempColor = BondColorMap_m[Index];
		}

	/*  Use default bond color  */
	else
		TempColor = BondColor_m;


	/*  Draw bond line  */
	if (TempColor!= CLEAR_COLOR)
		GR_Line
			(
			ImageCur_m[HOR][FirstIndex ],
			ImageCur_m[VER][FirstIndex ],
			ImageCur_m[HOR][SecondIndex],
			ImageCur_m[VER][SecondIndex],
			TempColor
			);
	}

void NormalizeSlipDisplacements
(
stepdata *CurStep,
unsigned *IxImageToCoord,
int		NumImagePart,
float 	SlipRadius,
int	  *ImageDisp[NDIR]
)
	{
	int image;

	/*  Return if Radius is zero	*/
	if (SlipRadius==0)
		return;


	/*  Set module-wide pointers to passed arrays  */
	ImageDisp_m[HOR] = ImageDisp[HOR];
	ImageDisp_m[VER] = ImageDisp[VER];

	/*  Allocate storage  */
	ALLOCATE (SumImageDisp_m[HOR], int, NumImagePart)
	ALLOCATE (SumImageDisp_m[VER], int, NumImagePart)
	ALLOCATE (CntImageDisp_m	, int, NumImagePart)

	/*  Call bond function	*/
	bondx
		(
		CurStep->c,
		IxImageToCoord,
		NumImagePart,
		DEFAULT_MIN_RADIUS,
		SlipRadius,
		CurStep->box,
		(unsigned *) CurStep->boundrep,
		SlipCallBack
		);

	/*  Modify displacements  */
	LOOP (image, NumImagePart)
		{

		if (CntImageDisp_m[image] != 0)
			{
			ImageDisp[HOR][image] -= SumImageDisp_m[HOR][image] /
				CntImageDisp_m[image];
			ImageDisp[VER][image] -= SumImageDisp_m[VER][image] /
				CntImageDisp_m[image];
			}
		}

	/*  Free storage	*/
	FREE (SumImageDisp_m[HOR])
	FREE (SumImageDisp_m[VER])
	FREE (CntImageDisp_m   )

	}



#pragma argsused
void SlipCallBack (int FirstIndex, int SecondIndex, float Radius)
	{
	SumImageDisp_m[HOR][FirstIndex ] += ImageDisp_m[HOR][SecondIndex];
	SumImageDisp_m[VER][FirstIndex ] += ImageDisp_m[VER][SecondIndex];
	CntImageDisp_m   [FirstIndex ] ++;
	SumImageDisp_m[HOR][SecondIndex] += ImageDisp_m[HOR][FirstIndex ];
	SumImageDisp_m[VER][SecondIndex] += ImageDisp_m[VER][FirstIndex ];
	CntImageDisp_m   [SecondIndex] ++;
	}




void CalcRotatedLimits
(
float RotatedCoordLimits[2][NDIR],
stepdata *CurStep,
dsparm *PlotParam
)
	{
	float CoordLimit		  [2] 		  [NDIR];
	float BoxCorner		  [NUM_CORNER][NDIR];
	float RotatedBoxCorner [NUM_CORNER][NDIR];
	float MinRange;
	float Range;
	int	idir;
	int	icorner;
	int	i;
	int	j;
	int	k;

	/*  Use fixed frame	*/
	if (PlotParam->FixedFrameFlag)
		{
		memcpy (CoordLimit, PlotParam->Frame, sizeof(CoordLimit));
		}

	/*  Use min/max particles	*/
	else
		{

		/*  Initialize Min/Max coordinates	*/
		CalcMinMaxCoord (CurStep, TRUE, TRUE);

		/*  
		If identical limits, redo Min/Max coordinates using ALL particles
		(not just selected particles)
		*/
		if (AreIdenticalLimits(CurStep->MinCoord, CurStep->MaxCoord))
			CalcMinMaxCoord (CurStep, FALSE, TRUE);

		/*  Copy limits  */
		LOOP (idir, NDIR)
			{
			CoordLimit[0][idir] = CurStep->MinCoord[idir];
			CoordLimit[1][idir] = CurStep->MaxCoord[idir];
			}

		/*  If limits still identical, use range from valid limits  */
		if (AreIdenticalLimits(CurStep->MinCoord, CurStep->MaxCoord))
			{
	
			/*  Find minimum range  */
			MinRange = 0;
			LOOP (idir, NDIR)
				{
				Range = CoordLimit[1][idir] - CoordLimit[0][idir];
				if (MinRange==0)
					MinRange = Range;
				else
					if (MinRange>Range && Range!=0.0)
						MinRange = Range;
				}

			/*  Set 0 ranges to minimum range  */
			LOOP (idir, NDIR)
				{
				if (CoordLimit[0][idir]==CoordLimit[1][idir])
					{
					CoordLimit[0][idir]  -= 0.5*MinRange;
					CoordLimit[1][idir]  += 0.5*MinRange;
					}
				}

			}


		}
		/*  Use min/max particles  */


	/*
	Find all 8 corners of box enclosing plotted atoms
	(NOTE:  BoxCorner[] could contain parallel-piped)
	*/
	icorner = 0;
	LOOP (i,2)
	LOOP (j,2)
	LOOP (k,2)
		{
		BoxCorner[icorner][X] = CoordLimit[i][X];
		BoxCorner[icorner][Y] = CoordLimit[j][Y];
		BoxCorner[icorner][Z] = CoordLimit[k][Z];
		icorner++;
		}

	/*  Rotate corners of box	*/

	LOOP (icorner, NUM_CORNER)
		{
		ROTATE_COORD (RotatedBoxCorner[icorner], PlotParam->Rotate, BoxCorner[icorner])
		}

	/*  Initialize limits  */
	memcpy
		(
		RotatedCoordLimits[0],
		RotatedBoxCorner[0],
		sizeof(RotatedCoordLimits[0])
		);
	memcpy
		(
		RotatedCoordLimits[1],
		RotatedBoxCorner[0],
		sizeof(RotatedCoordLimits[0])
		);

	/*  Search for min and max limits  */
	LOOP (icorner, NUM_CORNER)
		{
		MIN(RotatedCoordLimits[0][X], RotatedBoxCorner[icorner][X])
		MIN(RotatedCoordLimits[0][Y], RotatedBoxCorner[icorner][Y])
		MIN(RotatedCoordLimits[0][Z], RotatedBoxCorner[icorner][Z])
		MAX(RotatedCoordLimits[1][X], RotatedBoxCorner[icorner][X])
		MAX(RotatedCoordLimits[1][Y], RotatedBoxCorner[icorner][Y])
		MAX(RotatedCoordLimits[1][Z], RotatedBoxCorner[icorner][Z])
		}

	}


BOOLEAN AreIdenticalLimits (float MinLimits[NDIR], float MaxLimits[NDIR])
	{
	BOOLEAN IdenticalLimits;
	int     idir;

	/*  Test for identical upper and lower limit  */
	IdenticalLimits = FALSE;
	LOOP (idir, NDIR)
		{
		IdenticalLimits = 
			(IdenticalLimits || MinLimits[idir]==MaxLimits[idir]);
		}

	return IdenticalLimits;
	}



/*
Assume position 1 is inside box which runs from 0 to Box[]
Move   position 2 inside box along line joining 1 and 2
*/
void ClipPoint
(
int x1,
int y1,
int *x2,
int *y2,
int xmin,
int ymin,
int xmax,
int ymax
)
   {
   float hDel;
   float vDel;
   float Slope;

   /*  Displacement between two positions  */
   hDel = (*x2) - x1;
   vDel = (*y2) - y1;

   /*  Two points are coincident - no need to clip  */
   if (hDel==0 && vDel==0)
      return;

   /*  Horizontal Line  */
   if (vDel==0.0)
      {
      if ((*x2) < xmin)
         (*x2) = xmin;
      else if ((*x2) > xmax)
         (*x2) = xmax;
      return ;
      }

   /*  Vertical Line  */
   if (hDel==0)
      {
      if ((*y2) < ymin)
         (*y2) = ymin;
      else if ((*y2) > ymax)
         (*y2) = ymax;
      return;
      }


   /*  General case  */
   Slope = vDel / hDel;

   if ((*x2) < xmin)
      {
      (*y2) -= Slope * (*x2 - xmin);
      (*x2)  = xmin;
      }
   else if ((*x2) > xmax)
      {
      (*y2) += Slope * (xmax - (*x2));
      (*x2)  = xmax;
      }

   if ((*y2) < ymin)
      {
      (*x2) -= (*y2 - ymin) / Slope;
      (*y2)  = ymin;
      }
   else if ((*y2) > ymax)
      {
      (*x2) += (ymax - (*y2)) / Slope;
      (*y2)  = ymax;
      }

   }
